% Datei: SPBuchsatz_Export.lua
%
%  Die hier definierte Methode
%  liest die angegebene SPBuchsatz *.tex-Datei
%  und erstellt eine gleichnamige *.rtf-Datei
%  im Rich Text Format,
%  die beispielsweise von Microsoft Word und LibreOffice/OpenOffice Writer
%  gelesen werden kann.
%
%
%  CAVEAT: File providing SPBuchsatz system code.


\begin{luacode*}

  local lpeg = require "lpeg"
  local lpeg_C  = lpeg.C
  local lpeg_Ct = lpeg.Ct
  local lpeg_R  = lpeg.R
  local lpeg_match  = lpeg.match

  local utf8_codes do
      local utf8_x  = lpeg_R"\128\191"
      local utf8_1  = lpeg_R"\000\127"
      local utf8_2  = lpeg_R"\194\223" * utf8_x
      local utf8_3  = lpeg_R"\224\239" * utf8_x * utf8_x
      local utf8_4  = lpeg_R"\240\244" * utf8_x * utf8_x * utf8_x
      local utf8    = utf8_1 + utf8_2 + utf8_3 + utf8_4
      local split   = lpeg_Ct (lpeg_C (utf8)^0) * -1

      utf8_codes = function (str)
          str = str and tostring (str)
          if not str then return end
          return lpeg_match (split, str)
      end
  end
  

  local globalBraceLevel = 0
  
  
  function escapedText (src, preserveSpecialCodes)
      local length = src:len ()
      if length==0 then
          return src
      end

      local retVal = ""
      local i=0
      local skipChars = 0
      local closingBracesToBePreserved = {}
      
      local codes = utf8_codes (src)
      
      if codes then
          local endIdx = #codes
          
          if (endIdx > 0) and (globalBraceLevel > 0) then
              
              local braceLevel = globalBraceLevel
              
              for j = 1, endIdx do
                
                  if (codes[j] == "{") then
                      braceLevel = braceLevel + 1
                  elseif (codes[j] == "}") then
                      braceLevel = braceLevel - 1
                    
                      if (braceLevel < globalBraceLevel) then

                          closingBracesToBePreserved[#closingBracesToBePreserved + 1] = j

                          globalBraceLevel = globalBraceLevel -1

                          table.sort (closingBracesToBePreserved)
                        
                          break;
                      end
                  end
              end
          end
          
          for i = 1, endIdx do
              if skipChars > 0 then
                  skipChars = skipChars - 1
                  goto endOfCodeLoop
              end
              
              local code = codes [i]
              
              -- if desired we preserve the RTF tokens "{\i …", "{\i0 …", "{\cs12 …" and their respective "}"
              if (preserveSpecialCodes) then
                  if (     (code == "{")
                       and (i+2 < endIdx)
                       and (codes[i+1] == "\\")
                       and (    (     (codes[i+2] == "i")
                                  and (                        (codes[i+3] == " ") 
                                        or ((i+3 < endIdx) and (codes[i+3] == "0") and (codes[i+4] == " "))
                                      )
                                )
                             or (     (i+5 < endIdx)
                                  and (codes[i+2] == "c")
                                  and (codes[i+3] == "s")
                                  and (codes[i+4] == "1")
                                  and (codes[i+5] == "2")
                                  and (codes[i+6] == " ")
                                )
                           )
                     )        
                  then
                      local spaceFound = false
                      local oldLevel   = 0 --globalBraceLevel
                      local braceLevel = 0 --globalBraceLevel
                      skipChars = -1
                      for j = i, endIdx do
                          
                          if not spaceFound then
                              retVal    = retVal..codes[j]
                              skipChars = skipChars+1
                          end
                        
                          if (codes[j] == "{") then
                              braceLevel = braceLevel + 1
                          elseif (codes[j] == "}") then
                              braceLevel = braceLevel - 1
                              
                              if (braceLevel == oldLevel) then
                              
                                  closingBracesToBePreserved[#closingBracesToBePreserved + 1] = j

                                  table.sort (closingBracesToBePreserved)
                                  
                                  break;
                              end
                          elseif (codes[j] == " ") then
                              spaceFound = true
                          end
                      end
                      
                      if (braceLevel > oldLevel) then
                          globalBraceLevel = globalBraceLevel + (braceLevel - oldLevel)
                      end
                      
                      goto endOfCodeLoop
                  end
              end
              
              local value = unicode.utf8.byte (code)
              local knownCode = nil

              if (#closingBracesToBePreserved > 0) and (closingBracesToBePreserved[1] == i) then
                  knownCode = code
                  table.remove (closingBracesToBePreserved, 1)
              else
                  if (i < endIdx) then
                      local code2 = codes [i+1]
                      local value2 = unicode.utf8.byte (code2)
                      
                      if (value == 92) then         -- "\"
                          if (value2 == 37)         -- "\%"
                          then
                              goto endOfCodeLoop
                          end
                          
                          if (value2 == 45)         -- "\-"
                          then
                              skipChars = 1
                              goto endOfCodeLoop
                          end
                        
                          if (value2 == 92)         -- "\\"
                          then
                              knownCode = "\\u"..value.."\\\'3f".."\\u"..value2.."\\\'3f"
                              skipChars = 1
                          elseif (value2 == 32)     -- "\ "
                          then
                              knownCode = "\\u"..value2.."\\\'3f"
                              skipChars = 1
                          elseif (value2 == 44)     -- "\,"
                          then
                              knownCode = "\\~"
                              skipChars = 1
                          end
                      end
                      if (value == 45) and (value2 == 45) then      -- "--"
                          knownCode = "\\"..string.char(39).."96"
                          skipChars = 1
                      end
                  end

                  if     (value == 0x27) then       -- ersetzt ' durch echten Apostroph
                      knownCode = "\\u8217\\'92"  
                  elseif (value == 0x7E) then       -- "~"
                      knownCode = "\\~"
                  end
              end
              
              if knownCode then
                  retVal = retVal..knownCode
              else
                  
                  if (value>31 and value<91) or (value>96 and value<123) then
                      retVal = retVal..code
                  else
                      retVal = retVal.."\\u"..value.."\\\'3f"
                  end
              end
              
              ::endOfCodeLoop::
          end
      end
      
      -- Leerstelle ist wichtig:
      local lineCommand = "\\line "
      local lineCommandLen = lineCommand:len ()
      
      local retValLenOld = retVal:len ()
      local retValHadTrailingSpaceBefore = 
        (     (retValLenOld > 0)
          and (retVal:sub (retValLenOld, retValLenOld) == " ") )

      retVal =
        retVal:gsub("%\\'e2\\'80\\'99",           "\\u8217\\'92")     -- ersetzt ’ durch echten Apostroph
              :gsub(" %\\u92\\'3f%\\u92\\'3f ",   lineCommand)
              :gsub(" %\\u92\\'3f%\\u92\\'3f",    lineCommand)
              :gsub("%\\u92\\'3f%\\u92\\'3f ",    lineCommand)
              :gsub("%\\u92\\'3f%\\u92\\'3f",     lineCommand)
              :gsub(" %\\u92\\'3fZeilenumbruch ", " ")       -- also NICHT \line einsetzen!
              :gsub("%\\u92\\'3fZeilenumbruch",   "")        -- also NICHT \line einsetzen!
              :gsub("%\\u92\\'3ftextbackslash ",  "\\\\")
              :gsub("%\\u92\\'3ftextbackslash",   "\\\\")
              :gsub("%\\u92\\'3fTilde ",          "~")
              :gsub("%\\u92\\'3fTilde",           "~")
              :gsub("e.%\\u92\\'3f/V.",           "e.V.")
              
      if not retValHadTrailingSpaceBefore then
        local retValLen = retVal:len ()
        if (retValLen > lineCommandLen-1) then
          local posLineCommand = findLast (retVal, lineCommand)
          if (posLineCommand > 0) and (posLineCommand == retValLen - lineCommandLen) then
            -- remove trailing space:
            retVal = retVal:sub (1, retValLen-1)
          end
        end
      end
      
      return retVal
  end

  
  function ltrim (str)
      local startOfValid = 1
      local endPos   = #str

      while (startOfValid < endPos and str:byte(startOfValid) == 32) do
          startOfValid = startOfValid + 1
      end

      if startOfValid > endPos then
          return ''
      else
          return str:sub(startOfValid)
      end
  end

  
  function gsplit(s,sep)
      return coroutine.wrap(function()
          if s == '' or sep == '' then
              coroutine.yield(s)
              return
          end
          local lasti = 1
          for v,i in s:gmatch('(.-)'..sep..'()') do
              coroutine.yield(v)
              lasti = i
          end
          coroutine.yield(s:sub(lasti))
      end)
  end


  function entferneToken (src, spToken)
      retVal = ""
      local parts = gsplit(src, spToken)
      for partRaw in parts do
          retVal = retVal..partRaw
      end
      return retVal
  end

  
  -- Ist keepRawText true, so bleibt der Returnwert unverändert, ansonsten wird er escaped.
  -- 
  -- Falls der erste Parameter optional (also in eckigen Klammern steht, oder fehlt),
  -- gib zuEntfernendeParameteranzahl als negative Zahl an,
  -- also -3 anstatt 3, falls zwei Muss-Parameter und (gegebenenfalls) ein optionaler Parameter zu entfernen sind.
  -- 
  -- Das Minuszeichen besagt also, dass der erste Parameter optional ist,
  -- der dann in [] stünde und ebenfalls zu entfernen wäre.
  -- 
  -- Beispiel: -1 bedeutet, es gibt keinen Muss-Parameter, könnte evtl. aber einen []-Parameter geben.
  -- 
  -- Ist ein rtfToken angegeben, so wird das gefundene spToken ersetzt die Klammer davor gerückt und eine Leerstelle eingesetzt.
  -- Beispiel:
  --      alt: \kursiv{…}
  --      neu: {\i …}
  --      
  -- Ist keepRawText false, wird das Ergebnis (mit Ausnahme des rtfToken) escaped.
  -- 
  -- Die letzte Angabe (danachZuErhaltenderParameter) kann true oder false sein: Falls true, wird nach den
  -- (evtl.) zu entfernenden Parametern ein weiterer Parameter angenommen. Dessen Inhalt bleibt erhalten, die Klammern werden entfernt.
  -- 
  function codiereSchalter (src, spToken, rtfToken, keepRawText, entferneSchalter, entferneInhalt, zuEntfernendeParameteranzahl, danachZuErhaltenderParameter, preserveSpecialCodes)
    
      local firstPosOfToken = findFirst (src, spToken)
      
      if (firstPosOfToken == 0) then
          if keepRawText then
              return src
          else
            local retValue = escapedText (src, preserveSpecialCodes)
            return retValue
          end
      end
      
      local insertRtfToken = (rtfToken:len () > 0)

      local retVal = ""

      local partIdx = 0

      local parts = gsplit(src, spToken)
      for partRaw in parts do
        
          partIdx = partIdx + 1
        
          local part = ltrim (partRaw)
          
          local length = part:len ()
          if length > 0 then

              if (partIdx == 1) and (firstPosOfToken > 1) then
                  -- We found a leading part: just some text before the first token.
                  if keepRawText then
                      retVal = part
                  else
                      retVal = escapedText (part, preserveSpecialCodes)
                  end
              else
                  local parametersToBeSkipped = 0
                  local firstParamIsOptional  = (zuEntfernendeParameteranzahl < 0)
                  if firstParamIsOptional then
                      parametersToBeSkipped = -1 * zuEntfernendeParameteranzahl
                  else
                      parametersToBeSkipped = zuEntfernendeParameteranzahl
                  end

                  local firstChar = part:sub (1,1)
                  local startsWithBrackets = firstParamIsOptional and (firstChar == "[")

                  if firstParamIsOptional and (not startsWithBrackets) then
                    parametersToBeSkipped = parametersToBeSkipped - 1
                  end
                  
                  if startsWithBrackets or (firstChar == "{") then
                    
                      local openingBrace = "{"
                      local closingBrace = "}"
                    
                      local posOpen = 1
                      local posTest = posOpen+1
                      
                      while (parametersToBeSkipped > 0) do
                        
                          if startsWithBrackets then
                            openingBrace = "["
                            closingBrace = "]"
                          end
                          
                          local posClosed = findClosingBraceExt (part, posTest, openingBrace, closingBrace)

                          if startsWithBrackets then
                            openingBrace = "{"
                            closingBrace = "}"
                            startsWithBrackets = false  -- Only the FIRST parameter uses [ ].
                          end
                          
                          if posClosed > 0 then

                              parametersToBeSkipped = parametersToBeSkipped - 1

                              if not entferneSchalter then
                                retVal = retVal..openingBrace
                              end
                              
                              if insertRtfToken then
                                  retVal = retVal..rtfToken.." "
                                  insertRtfToken = false
                              end

                              if not entferneSchalter then
                                  if keepRawText then
                                      retVal =              retVal..part:sub (posOpen+1, posClosed)
                                  else
                                      retVal = retVal..escapedText (part:sub (posOpen+1, posClosed), preserveSpecialCodes)
                                  end
                              end
                              
                              posTest = posClosed + 1
                              local posNextOpen = findFrom (part, openingBrace, posTest)
                              if (posNextOpen > 0) and (posNextOpen < length) then
                                  posTest = posNextOpen+1
                                  posOpen = posNextOpen
                              else
                                  parametersToBeSkipped = 0
                              end     

                          else
                              retVal = retVal.." \\"..openingBrace.."…"..closingBrace.."? "  -- we export \{ … \} to indicate what's wrong
                              parametersToBeSkipped = 0
                          end

                      end -- of while

                      -- process last parameter and/or the remaining text of the part:
                      if danachZuErhaltenderParameter and (posOpen > 0) then
                          local posClosed = findClosingBraceExt (part, posTest, openingBrace, closingBrace)
                          if (posClosed > 0) then
                              
                              if insertRtfToken then
                                  retVal = retVal..openingBrace..rtfToken.." "
                              end
                              
                              if keepRawText then
                                  retVal =              retVal..part:sub (posOpen+1, posClosed-1)
                              else
                                  retVal = retVal..escapedText (part:sub (posOpen+1, posClosed-1), preserveSpecialCodes)
                              end
                              
                              if insertRtfToken then
                                  retVal = retVal..closingBrace
                                  insertRtfToken = false
                              end
                              
                              if keepRawText then
                                  retVal =              retVal..part:sub (posClosed+1)
                              else
                                  retVal = retVal..escapedText (part:sub (posClosed+1), preserveSpecialCodes)
                              end
                          else
                              if keepRawText then
                                  retVal =              retVal..part:sub (posTest)
                              else
                                  retVal = retVal..escapedText (part:sub (posTest), preserveSpecialCodes)
                              end
                          end
                      else
                          if keepRawText then
                              retVal =              retVal..part:sub (posTest)
                          else
                              retVal = retVal..escapedText (part:sub (posTest), preserveSpecialCodes)
                          end
                      end
                  else
                      if keepRawText then
                          retVal = retVal..part
                      else
                          retVal = retVal..escapedText (part, preserveSpecialCodes)
                      end
                  end
              end
          end
      end

      return retVal
  end

  
  function kommentarEinbetten (Datei, SPBuchsatzVersion, Text)
      Datei:write("{\n{\\*\\atnid SPB}{\\*\\atnauthor SPBuchsatz "..SPBuchsatzVersion.." }\\chatn{\\*\\annotation{\\*\\atndate 0}\\cf1\\kerning1\\dbch\\af6\\langfe1031\\ltrpar\\aspalpha\\li0\\ri0\\lin0\\rin0\\fi0\\sb0\\sa0\\sl240\\slmult1\\ql\\rtlch \\ltrch\\loch\\fs24\\lang1031{\\expnd0\\expndtw0\\b0\\ab0\\dbch\\af8\\shad0\\kerning1\\scaps0\\caps0\\charscalex100\\outl0\\i0\\ai0\\afs24\\langfe1033\\alang1081\\cf0\\dbch\\af9\\accnone\\ulnone\\ulc0\\strike0\\fs20\\loch\\f3\\lang1033\n"..Text.."\\uc1 }}}")
  end
  
  
  spRtfTagNewPage = "\\pagebb"
  
  
  function spExportToRTF (srcFileName, destFileName, SPBuchsatzVersion, defaultSceneDividerIsText, defaultSceneDividerIsImage)
  
      globalBraceLevel = 0
      
      print ("Datei wird gelesen: "..srcFileName.."\nDatei wird erzeugt: "..destFileName)
      io.write ("[")

      local rtfDocument = io.open(destFileName,"w")
      
      rtfDocument:write("{\\rtf1\\ansi\\deff3\\adeflang1025\n")
      
      rtfDocument:write("{\\*\\generator SPBuchsatz/"..SPBuchsatzVersion.."}\n")

      local dt = os.date("*t")
      rtfDocument:write("{\\info")
            rtfDocument:write("{\\title SPBuchsatz-Text-Transferdatei}")
            rtfDocument:write("{\\doccomm Diese Datei wurde erzeugt von SPBuchsatz "..SPBuchsatzVersion.."\\u16\\'10\\u10\\'0anach dem Textsatz am "..dt.day.."."..dt.month.."."..dt.year..", um "..dt.hour..":"..dt.min.." Uhr.\\u16\\'10\\u10\\'0a\\u16\\'10\\u10\\'0aLade sie in Dein Text-Schreibprogramm\\u16\\'10\\u10\\'0aals Grundlage zur Erzeugung der eBook-Version.}")
            rtfDocument:write("{\\creatim\\yr"..dt.year.."\\mo"..dt.month.."\\dy"..dt.day.."\\hr"..dt.hour.."\\min"..dt.min.."}")
            rtfDocument:write("{\\author SPBuchsatz "..SPBuchsatzVersion.." }")
            rtfDocument:write("{\\revtim\\yr"..dt.year.."\\mo"..dt.month.."\\dy"..dt.day.."\\hr"..dt.hour.."\\min"..dt.min.."}")
            rtfDocument:write("{\\printim\\yr0\\mo0\\dy0\\hr0\\min0}\n")
      rtfDocument:write("}\n")
      
      rtfDocument:write("{\\fonttbl\n")
          rtfDocument:write("{\\f0\\froman\\fprq2\\fcharset0 Times New Roman;}\n")
          rtfDocument:write("{\\f1\\froman\\fprq2\\fcharset2 Symbol;}\n")
          rtfDocument:write("{\\f2\\fswiss\\fprq2\\fcharset0 Arial;}\n")
          rtfDocument:write("{\\f3\\froman\\fprq2\\fcharset0 Liberation Serif{\\*\\falt Times New Roman};}\n")
          rtfDocument:write("{\\f4\\froman\\fprq2\\fcharset0 Times New Roman;}\n")
          rtfDocument:write("{\\f5\\fswiss\\fprq2\\fcharset0 Times New Roman;}\n")
          rtfDocument:write("{\\f6\\fnil\\fprq2\\fcharset0 Times New Roman;}\n")
      rtfDocument:write("}\n")
      
      local BlocksatzCode = "\\qj"
      local LombardenCode = "\\afs32\\fs32"
      
      rtfDocument:write("{\\stylesheet\n")
      rtfDocument:write("{\\*\\cs10\\cf0 \\additive Default Paragraph Font;}\n")
      rtfDocument:write("{\\*\\cs12 \\additive "..LombardenCode.." \\sbasedon10 Lombarde Size;}\n")
      rtfDocument:write("{\\s0\\snext0\\ql\\widctlpar\\hyphpar0\\ltrpar\\cf1\\kerning1\\dbch\\af6\\lang1031\\langfe1031\\dbch\\af6\\afs24\\alang1031\\loch\\f4\\fs24\\lang1031 Normal;}\n")
          
          
      -- Besonderheit: Wir wollen Buchteile und Kapitel im Inhaltsverzeichnis des eBuchs haben, dehalb erhalten beide(!) die höchste Gliederungsebene (outlinelevel: 0), obwohl sie ja eigentlich unterschiedliche Ebenen haben.
          
      rtfDocument:write("{\\s1\\sbasedon15\\snext1\\outlinelevel0\\qc\\widctlpar\\hyphpar0\\sb1191\\sa0\\keepn\\ltrpar\\cf1\\b0\\kerning1\\dbch\\af4\\langfe1031\\dbch\\af8\\afs24\\loch\\f5\\fs56\\lang1031 Part;}\n")
      rtfDocument:write("{\\s2\\sbasedon15\\snext2\\outlinelevel0\\qc\\widctlpar\\hyphpar0\\sb198\\sa510\\keepn\\ltrpar\\cf1\\b0\\kerning1\\dbch\\af6\\langfe1031\\dbch\\af6\\afs24\\alang1031\\loch\\f5\\fs42\\lang1031 Chapter;}\n")
          
          
      rtfDocument:write("{\\s15\\sbasedon0\\snext16\\ql\\widctlpar\\hyphpar0\\sb240\\sa120\\keepn\\ltrpar\\cf1\\kerning1\\dbch\\af6\\langfe1031\\dbch\\af6\\afs24\\alang1031\\loch\\f4\\fs28\\lang1031 Heading;}\n")
      rtfDocument:write("{\\s16\\sbasedon0\\snext16\\sl276\\slmult1\\ql\\widctlpar\\hyphpar0"..BlocksatzCode.."\\sb0\\sa0\\ltrpar\\cf1\\kerning1\\dbch\\af6\\langfe1031\\dbch\\af6\\afs24\\alang1031\\loch\\f4\\fs24\\lang1031 Text Body;}\n")
      rtfDocument:write("{\\s17\\sbasedon16\\snext17\\sl276\\slmult1\\ql\\widctlpar\\hyphpar0\\sb0\\sa0\\ltrpar\\cf1\\kerning1\\dbch\\af6\\langfe1031\\dbch\\af6\\afs24\\alang1031\\loch\\f4\\fs24\\lang1031 List;}\n")
      rtfDocument:write("{\\s18\\sbasedon0\\snext18\\ql\\widctlpar\\hyphpar0\\sb120\\sa120\\ltrpar\\cf1\\i\\kerning1\\dbch\\af6\\langfe1031\\dbch\\af6\\afs24\\alang1031\\loch\\f4\\fs24\\lang1031 Caption;}\n")
      rtfDocument:write("{\\s19\\sbasedon0\\snext19\\ql\\widctlpar\\hyphpar0\\ltrpar\\cf1\\kerning1\\dbch\\af6\\langfe1031\\dbch\\af6\\afs24\\alang1031\\loch\\f4\\fs24\\lang1031 Index;}\n")
      
      rtfDocument:write("{\\s20\\sbasedon16\\snext16\\sl276\\slmult1\\qc\\widctlpar\\hyphpar0\\sb340\\sa340\\ltrpar\\cf1\\kerning1\\dbch\\af6\\langfe1031\\dbch\\af6\\afs24\\alang1031\\loch\\f4\\fs24\\lang1031 SzenentrennerMitTextOderBild;}\n")
      
      rtfDocument:write("{\\s21\\sbasedon16\\snext21\\sl276\\slmult1\\ql\\widctlpar\\hyphpar0\\li283\\ri0\\lin283\\rin0\\fi0"..BlocksatzCode.."\\sb0\\sa0\\ltrpar\\cf1\\kerning1\\dbch\\af6\\langfe1031\\dbch\\af6\\afs24\\alang1031\\loch\\f4\\fs24\\lang1031 Text Body Indent;}\n")
      rtfDocument:write("{\\s22\\sbasedon16\\snext22\\sl276\\slmult1\\ql\\widctlpar\\hyphpar0\\li0\\ri0\\lin0\\rin0\\fi397"..BlocksatzCode.."\\sb0\\sa0\\ltrpar\\cf1\\kerning1\\dbch\\af6\\langfe1031\\dbch\\af6\\afs24\\alang1031\\loch\\f4\\fs24\\lang1031 Erstzeileneinzug;}\n")
      rtfDocument:write("{\\s23\\sbasedon16\\snext16\\sl276\\slmult1\\ql\\widctlpar\\hyphpar0\\sb0\\sa0\\ltrpar\\cf1\\kerning1\\dbch\\af4\\langfe1031\\dbch\\af4\\afs24\\loch\\f4\\fs42\\ KapitelTitelSchrift;}\n")
      rtfDocument:write("{\\s24\\sbasedon16\\snext25\\sl276\\slmult1\\ql\\widctlpar\\hyphpar0\\li397\\ri397\\lin397\\rin397\\fi0"..BlocksatzCode.."\\sb283\\sa0\\contextualspace\\ltrpar\\cf1\\kerning1\\dbch\\af4\\langfe1031\\dbch\\af8\\afs24\\loch\\f4\\fs24\\lang1031 Einrueckung;}\n")
      rtfDocument:write("{\\s25\\sbasedon24\\snext25\\sl276\\slmult1\\ql\\widctlpar\\hyphpar0\\li397\\ri397\\lin397\\rin397\\fi397"..BlocksatzCode.."\\sb0\\sa283\\contextualspace\\ltrpar\\cf1\\kerning1\\dbch\\af4\\langfe1031\\dbch\\af8\\afs24\\loch\\f4\\fs24\\lang1031 Einrueckung_und_Erstzeileneinzug;}\n")

      rtfDocument:write("{\\s26\\sbasedon16\\snext16\\sl276\\slmult1\\qc\\widctlpar\\hyphpar0\\sb170\\sa0\\ltrpar\\cf1\\kerning1\\dbch\\af6\\langfe1031\\dbch\\af6\\afs24\\alang1031\\loch\\f4\\fs24\\lang1031 SzenentrennerLeer;}\n")
      
      rtfDocument:write("}\n")

      rtfDocument:write("\\pgndec\n")

      local iLine = 0
      
      local iKapitel = 0
      
      local inKursivemBereich = false
      local inZentriertemBereich = false
      local inEingeruecktemBereich = false
      local AnzahlEingerueckterAbsaetze = 0
      local neueSeiteAngefordert = false
      local erzwingeKeinEinzug = false
      
      local buchteilKapitelCodes = "{\\rtlch \\ltrch\\loch\n"
      
      local absatzAbschlussCodes = "}\\par\n"
      
      local rawFile = io.open (srcFileName, "r")
      for lineIn in rawFile:lines () do
          local lineStripped = lineIn:gsub(" ", "")
          if lineStripped:len () > 0 then
              
              if (lineStripped == "\\DateiHierBeenden") then
                  break;
              end
              
              if not startsWith (lineIn, "%") then
                  
                  if hasToken (lineStripped, "\\BeginnKursiv") then
                      inKursivemBereich = true
                      goto endOfLineLoop
                  end
                  if hasToken (lineStripped, "\\EndeKursiv") then
                      inKursivemBereich = false
                      goto endOfLineLoop
                  end
              
                  if hasToken (lineStripped, "\\OhneSeitenzahl") or
                     hasToken (lineStripped, "\\KeineSeitenzahl") or
                     hasToken (lineStripped, "\\MitSeitenzahl") or
                     hasToken (lineStripped, "\\AbstandImIV") or
                     hasToken (lineStripped, "\\setzeInhaltsverzeichnis") or
                     hasToken (lineStripped, "\\enlargethispage") or
                     hasToken (lineStripped, "\\HalbeLeerzeileOben") or
                     hasToken (lineStripped, "\\GeheZuRegisterzeile") or
                     hasToken (lineStripped, "\\BeginnWortabstand{") or
                     hasToken (lineStripped, "\\BeginnLaufweite{") or
                     hasToken (lineStripped, "\\BeginnEng") or
                     hasToken (lineStripped, "\\BeginnHalbEng") or
                     hasToken (lineStripped, "\\BeginnHalbWeit") or
                     hasToken (lineStripped, "\\BeginnWeit") or
                     hasToken (lineStripped, "\\EndeLaufweite") or
                     hasToken (lineStripped, "\\EndeWortabstand")
                  then
                      goto endOfLineLoop
                  end

                  if hasToken (lineStripped, "\\neueSeite") then
                      neueSeiteAngefordert = true
                      goto endOfLineLoop
                  end


                  local lineOut = lineIn:gsub("%\\Leerschritt", " ") -- remove all occurences of \Leerschritt by a normal space character
                                                                     -- (these "\Leerschritt" were former "\ " strings, replaced by SPBuchsatz preprocessing for further TeX processing, so we are reverting that replacing here)
                  lineOut = strSimplified (lineOut)
                  
                  lineOut = lineOut:gsub("%\\SPBuchsatz ",  "SPBuchsatz ")
                                   :gsub("%\\SPBuchsatz%,", "SPBuchsatz,")
                                   :gsub("%\\SPBuchsatz%;", "SPBuchsatz;")
                                   :gsub("%\\SPBuchsatz%.", "SPBuchsatz.")
                                   :gsub("%\\SPBuchsatz%!", "SPBuchsatz!")
                                   :gsub("%\\SPBuchsatz%?", "SPBuchsatz?")
                    
                    
                  local posGeschichtenKapitel = findFirst (lineOut, "\\GeschichtenKapitel")
                  if posGeschichtenKapitel == 0 then
                    posGeschichtenKapitel = findFirst (lineOut, "\\Geschichtenkapitel")
                  end
                  if posGeschichtenKapitel == 0 then
                    posGeschichtenKapitel = findFirst (lineOut, "\\ZwischenTitel")
                  end
                  if posGeschichtenKapitel == 0 then
                    posGeschichtenKapitel = findFirst (lineOut, "\\Zwischentitel")
                  end
                  if posGeschichtenKapitel > 0 then
                    local posOpen = findAfter (lineOut, '{', posGeschichtenKapitel)
                    if (posOpen > 0) then
                      local posClose = findClosingBraceExt (lineOut, posOpen+1, '{', '}')
                      if (posClose > 0) then
                       
                       rtfDocument:write ("\\line")  -- we set an additional empty line before an inner sub title
                       
                       lineOut = "\\Beitext{\\Normalschrift{\\Lombarde{"..ltrim (lineOut:sub (posOpen+1, posClose-1)).."}}}"

                       lineStripped = lineOut:gsub(" ", "")
                      end
                    end
                  end
                  
                  
                  lineOut = codiereSchalter (lineOut, "\\Notiz",             "", true, true, true,  0, false, false)
                  lineOut = codiereSchalter (lineOut, "\\Warnung",           "", true, true, true,  0, false, false)

                  lineOut = codiereSchalter (lineOut, "\\mbox",              "", true, true, true,  0, false, false)

                  lineOut = codiereSchalter (lineOut, "\\altneu",            "", true, true, false, 1, true, false)
                  if hasToken (lineStripped, "\\alt{")
                  then
                    lineOut=codiereSchalter (lineOut, "\\alt",               "", true, true, true,  0, true, false)
                  end
                  if hasToken (lineStripped, "\\neu{")
                  then
                    lineOut=codiereSchalter (lineOut, "\\neu",               "", true, true, false, 0, false, false)
                  end
                  lineOut = codiereSchalter (lineOut, "\\Bruchzahl",         "", true, true, false, 0, true, false)
                  lineOut = codiereSchalter (lineOut, "\\normaleZiffern",    "", true, true, false, 0, true, false)

                  --lineOut = codiereSchalter (lineOut, "\\Lombarde",          "", true, true, false, 0, true, false)
                  lineOut = codiereSchalter (lineOut, "\\Initial"  ,         "", true, true, false,-1, true, false)

                  lineOut = codiereSchalter (lineOut, "\\inlineWortabstand", "", true, true, false, 1, true, false)
                  lineOut = codiereSchalter (lineOut, "\\inlineLaufweite",   "", true, true, false,-2, true, false)  -- first parameter is optional, thus -2 instead of 2
                  
                  lineOut = codiereSchalter (lineOut, "\\inlineEtwasWeiter", "", true, true, false, 0, true, false)
                  lineOut = codiereSchalter (lineOut, "\\inlineEtwasWeit",   "", true, true, false, 0, true, false)
                  lineOut = codiereSchalter (lineOut, "\\inlineEtwasEnger",  "", true, true, false, 0, true, false)
                  lineOut = codiereSchalter (lineOut, "\\inlineEtwasEng",    "", true, true, false, 0, true, false)
                  lineOut = codiereSchalter (lineOut, "\\inlineHalbWeit",    "", true, true, false, 0, true, false)
                  lineOut = codiereSchalter (lineOut, "\\inlineHalbEng",     "", true, true, false, 0, true, false)
                  lineOut = codiereSchalter (lineOut, "\\inlineWeitest",     "", true, true, false, 0, true, false)
                  lineOut = codiereSchalter (lineOut, "\\inlineWeiter",      "", true, true, false, 0, true, false)
                  lineOut = codiereSchalter (lineOut, "\\inlineWeit",        "", true, true, false, 0, true, false)
                  lineOut = codiereSchalter (lineOut, "\\inlineEnger",       "", true, true, false, 0, true, false)
                  lineOut = codiereSchalter (lineOut, "\\inlineEngst",       "", true, true, false, 0, true, false)
                  lineOut = codiereSchalter (lineOut, "\\inlineEng",         "", true, true, false, 0, true, false)
                  
                  lineOut = entferneToken   (lineOut, "\\keinAbs")
                  lineOut = entferneToken   (lineOut, "\\neuerAbs")
                  lineOut = entferneToken   (lineOut, "\\trenn\\Zeilenumbruch ")
                  lineOut = entferneToken   (lineOut, "\\trenn ")
                  lineOut = entferneToken   (lineOut, "\\trenn")
                  lineOut = entferneToken   (lineOut, "\\keineLombarde ")
                  lineOut = entferneToken   (lineOut, "\\keineLombarde")
                  lineOut = entferneToken   (lineOut, "\\keinInitial ")
                  lineOut = entferneToken   (lineOut, "\\keinInitial")
                  
                  
                  local besondereBildSeite = hasToken (lineStripped, "\\neueBildseite")
                  if besondereBildSeite then
                      neueSeiteAngefordert = true
                  end
                  
                  local posBeginnEinrueckung = findFirst (lineOut, "\\BeginnEinrueckung") 
                  if posBeginnEinrueckung > 0 then
                      inEingeruecktemBereich = true
                      local Restzeile = ltrim (lineOut:sub (posBeginnEinrueckung+18))
                      if Restzeile:sub(1, 1) == "*" then
                          Restzeile = Restzeile:sub(2)
                      end
                      if Restzeile:gsub(" ", ""):len () == 0 then
                          goto endOfLineLoop
                      end
                      erzwingeKeinEinzug = true
                      lineOut = Restzeile
                  end
                  if hasToken (lineStripped, "\\EndeEinrueckung") then
                      inEingeruecktemBereich = false
                      if AnzahlEingerueckterAbsaetze > 1 then
                          goto endOfLineLoop
                      end
                      lineOut = ""
                  end

                  local posBeginnZentrierung = findFirst (lineOut, "\\BeginnZentriert")
                  if posBeginnZentrierung > 0 then
                      inZentriertemBereich = true
                      local Restzeile = ltrim (lineOut:sub (posBeginnZentrierung+16))
                      if Restzeile:sub(1, 1) == "*" then
                          Restzeile = Restzeile:sub(2)
                      else
                        rtfDocument:write ("\\line")
                      end
                      if Restzeile:gsub(" ", ""):len () == 0 then
                          goto endOfLineLoop
                      end
                      erzwingeKeinEinzug = true
                      lineOut = Restzeile
                  end
                  local posEndeZentrierung = findFirst (lineOut, "\\EndeZentriert") 
                  if posEndeZentrierung > 0 then
                      inZentriertemBereich = false
                      local Restzeile = ltrim (lineOut:sub (posEndeZentrierung+14))
                      if Restzeile:sub(1, 1) ~= "*" then
                        rtfDocument:write ("\\line")
                      end
                      goto endOfLineLoop
                  end
                  
                  iLine = iLine+1
                  if iLine < 10 or iLine % 10 == 0 then
                    io.write (".")
                  end
                  
                  rtfDocument:write("\\pard\\plain")

                  local eventuellNeueSeiteCode = ""
                  if neueSeiteAngefordert then
                      neueSeiteAngefordert = false
                      eventuellNeueSeiteCode = spRtfTagNewPage
                  end

                  local extraBuchteilZeileVorab = nil
                  
                  local hasBuchteil = hasToken (lineStripped, "\\Buchteil{")
                  if hasBuchteil then
                      local posOpen2   = findLast (lineOut, "{")
                      local posClosed2 = findLast (lineOut, "}")
                      if (posOpen2 > 0) and (posOpen2 < posClosed2) then
                          local posOpen1   = findSecondLast (lineOut, "{")
                          local posClosed1 = findSecondLast (lineOut, "}")
                          if (posOpen1 > 0) and (posOpen1 < posClosed1) then
                              extraBuchteilZeileVorab = lineOut:sub(posOpen1+1, posClosed1-1)
                          end
                          lineOut = lineOut:sub(posOpen2+1, posClosed2-1)
                      else
                          lineOut = lineOut:gsub("%\\Buchteil","")
                      end
                  end
                  
                  local foundShortStory = false
                  
                  local posKapitel = findFirst (lineStripped, "\\Kapitel{")
                  if posKapitel==0 then
                      posKapitel = findFirst (lineStripped, "\\Kapitel[")
                  end
                  if posKapitel==0 then
                      posKapitel = findFirst (lineStripped, "\\KapitelSpezial{")
                  end
                  if posKapitel==0 then
                      posKapitel = findFirst (lineStripped, "\\KapitelSpezial[")
                  end
                  if posKapitel==0 then
                      posKapitel = findFirst (lineStripped, "\\Prolog{")
                  end
                  if posKapitel==0 then
                      posKapitel = findFirst (lineStripped, "\\Prolog[")
                  end
                  if posKapitel==0 then
                      posKapitel = findFirst (lineStripped, "\\PrologEinfach{")
                  end
                  if posKapitel==0 then
                      posKapitel = findFirst (lineStripped, "\\PrologEinfach[")
                  end
                  if posKapitel==0 then
                      posKapitel = findFirst (lineStripped, "\\PrologOhneHintergrundBild{")
                  end
                  if posKapitel==0 then
                      posKapitel = findFirst (lineStripped, "\\PrologOhneHintergrundBild[")
                  end
                  if posKapitel==0 then
                      posKapitel = findFirst (lineStripped, "\\Epilog{")
                  end
                  if posKapitel==0 then
                      posKapitel = findFirst (lineStripped, "\\Epilog[")
                  end
                  if posKapitel==0 then
                      posKapitel = findFirst (lineStripped, "\\EpilogEinfach{")
                  end
                  if posKapitel==0 then
                      posKapitel = findFirst (lineStripped, "\\EpilogEinfach[")
                  end
                  if posKapitel==0 then
                      posKapitel = findFirst (lineStripped, "\\EpilogOhneHintergrundBild{")
                  end
                  if posKapitel==0 then
                      posKapitel = findFirst (lineStripped, "\\EpilogOhneHintergrundBild[")
                  end
                  if posKapitel==0 then
                    posKapitel = findFirst (lineStripped, "\\Geschichte{")
                    if posKapitel then
                        foundShortStory = true
                    end
                  end
                  if posKapitel==0 then
                    posKapitel = findFirst (lineStripped, "\\Geschichte[")
                    if posKapitel > 0 then
                        foundShortStory = true
                    end
                  end
                  if posKapitel==0 then
                    posKapitel = findFirst (lineStripped, "\\GeschichteLinks{")
                    if posKapitel then
                        foundShortStory = true
                    end
                  end
                  if posKapitel==0 then
                    posKapitel = findFirst (lineStripped, "\\GeschichteLinks[")
                    if posKapitel > 0 then
                        foundShortStory = true
                    end
                  end
                  
                  local posKapitelTitelSchrift = findFirst (lineOut, "{\\KapitelTitelSchrift ")
                  local hasKapitelFortlaufend  = startsWith (lineOut, "\\KapitelFortlaufend")

                  local posBeitext = findFirst (lineStripped, "\\Beitext")
                  if (posBeitext > 0) then

                    rtfDocument:write ("\\line\\line")
                    
                    local posFirstParam = posBeitext + 8
                    local hasBeitextMitAbstand = hasToken (lineStripped, "\\BeitextMitAbstand")
                    if hasBeitextMitAbstand then
                      posFirstParam = posBeitext + 18
                    end

                    local params = ltrim (lineStripped:sub (posFirstParam))

                    if (params:sub (1, 1) ~= "[") or
                       (params:sub (2, 2) == "]")
                    then
                      inZentriertemBereich = true -- no [] parameter, or [] value missing
                    else
                      kommentarEinbetten (rtfDocument, SPBuchsatzVersion,
                                          "Justierung der Zeile korrigieren:\\par Linksbuendig, rechtsbuendig, Blocksatz?")
                    end

                    inKursivemBereich = true

                    if hasBeitextMitAbstand then
                      lineOut = codiereSchalter (lineOut, "\\BeitextMitAbstand", "", true, true, false, -3, true, false)
                    else
                      lineOut = codiereSchalter (lineOut, "\\Beitext",           "", true, true, false, -1, true, false)
                    end
                  end
                  
                  local imageWanted = false
                  
                  if hasBuchteil or (posKapitel > 0) or (posKapitelTitelSchrift > 0) or hasKapitelFortlaufend then
                      erzwingeKeinEinzug = true
                      
                      if (posKapitel > 0) or hasKapitelFortlaufend then
                          iKapitel = iKapitel + 1
                      end
                      
                      if hasBuchteil or (posKapitel > 0) then
                          eventuellNeueSeiteCode = spRtfTagNewPage
                          
                          local posBracket = findAfter (lineOut, '[', posKapitel)
                          local posOpen    = findAfter (lineOut, '{', posKapitel)

                          if (posBracket > 0) and (posBracket < posOpen) then
                              posBracket = findClosingBraceExt (lineOut, posBracket+1, '[', ']')
                              posOpen = findAfter (lineOut, '{', posBracket)
                          end

                          if hasBuchteil then
                              local posTest = findClosingBraceExt (lineOut, posOpen+1, '{', '}')  -- we skip the first parameter
                              posOpen = findAfter (lineOut, '{', posTest)
                          end

                          local posClosed = findLast (lineOut, "}")

                          if (posOpen > 0) and (posOpen < posClosed) then

                              lineOut = ltrim (lineOut:sub (posOpen+1, posClosed-1))

                              if hasBuchteil then
                                  lineOut = lineOut:gsub ("}{", "\\line\\line" , 1)  -- we separate the 2nd and 3rd parameter by an empty line
                              end

                              if foundShortStory then
                                  lineOut = lineOut:gsub ("}{", " " , 1)
                                  lineOut = lineOut:gsub ("}{", ": ", 1)
                              end
                          end
                          
                          if hasBuchteil then
                              rtfDocument:write("\\s1\\qc\\widctlpar\\hyphpar0\\sb1191\\sa0\\keepn\\ltrpar\\cf1\\b0\\kerning1\\dbch\\af4\\langfe1031"..eventuellNeueSeiteCode.."\\dbch\\af8\\afs24\\loch\\f5\\fs56\\lang1031")
                          else
                              rtfDocument:write("\\s2\\qc\\widctlpar\\hyphpar0\\sb198\\sa510\\keepn\\ltrpar\\cf1\\b0\\kerning1\\dbch\\af6\\langfe1031"..eventuellNeueSeiteCode.."\\dbch\\af6\\afs24\\alang1031\\loch\\f5\\fs42\\lang1031")
                          end
                      else
                          if hasKapitelFortlaufend then
                              rtfDocument:write("\\s23\\sl276\\slmult1\\ql\\widctlpar\\hyphpar0\\sb0\\sa0\\ltrpar\\cf1\\kerning1\\dbch\\af4\\langfe1031\\dbch\\af4\\afs24\\loch\\f4\\fs42\\lang1031\\qc")
                              lineOut = "• "..iKapitel.." •"
                          else
                              local posStart  = findAfter (lineOut, " ", posKapitelTitelSchrift)
                              local posClosed = findAfter (lineOut, "}", posStart)
                              
                              if (posStart > 0) and (posClosed > posStart) then
                                  lineOut = ltrim(lineOut:sub(posStart+1, posClosed-1))

                              end
                              rtfDocument:write("\\s23\\sl276\\slmult1\\ql\\widctlpar\\hyphpar0\\sb0\\sa0\\ltrpar\\cf1\\kerning1\\dbch\\af4\\langfe1031\\dbch\\af4\\afs24\\loch\\f4\\fs42\\lang1031"..eventuellNeueSeiteCode)
                          end
                      end
                      rtfDocument:write (buchteilKapitelCodes)
                  else
                      if hasToken (lineStripped, "\\Szenentrenner", 1, true) then
                          erzwingeKeinEinzug = true
                          
                          local textWanted = not hasToken (lineStripped, "\\SzenentrennerLeer", 1, true)  -- (actually this tests if text is PERHAPS wanted)
                          
                          if textWanted then
                            textWanted  = hasToken (lineStripped, "\\SzenentrennerText", 1, true)
                            imageWanted = hasToken (lineStripped, "\\SzenentrennerBild", 1, true)
                            
                            if not (textWanted or imageWanted) then
                                textWanted = (defaultSceneDividerIsText:len () > 0)
                                if not textWanted then
                                  imageWanted = (defaultSceneDividerIsImage:len () > 0)
                                end
                            end
                          end
                          
                          if (textWanted or imageWanted) then
                            rtfDocument:write("\\s20\\sl276\\slmult1\\qc\\widctlpar\\hyphpar0\\sb340\\sa340\\ltrpar\\cf1\\kerning1\\dbch\\af6\\langfe1031"..eventuellNeueSeiteCode.."\\dbch\\af6\\afs24\\alang1031\\loch\\f4\\fs24\\sl276\\slmult1\\li0\\ri0\\lin0\\rin0\\fi0")
                            if textWanted then
                              lineOut = "****"
                            else
                              lineOut = "(Bitte setze hier das Szenentrenner-Bild ein.)"
                            end
                          else
                            rtfDocument:write("\\s26\\sl276\\slmult1\\qc\\widctlpar\\hyphpar0\\sb170\\sa0\\ltrpar\\cf1\\kerning1\\dbch\\af6\\langfe1031"..eventuellNeueSeiteCode.."\\dbch\\af6\\afs24\\alang1031\\loch\\f4\\fs24\\sl276\\slmult1\\li0\\ri0\\lin0\\rin0\\fi0")
                            lineOut = " "
                          end
                      else
                          if posBeitext > 0 then
                              erzwingeKeinEinzug = true
                          end
                          
                          local kursiverBereich = ""
                          if inKursivemBereich then
                              kursiverBereich = "\\i"
                          end
                          
                          if inEingeruecktemBereich then
                              AnzahlEingerueckterAbsaetze = AnzahlEingerueckterAbsaetze + 1
                          else
                              AnzahlEingerueckterAbsaetze = 0
                          end
                          
                          local BlocksatzCode
                          if inZentriertemBereich then
                              BlocksatzCode = ""
                          else
                              BlocksatzCode = "\\qj"
                          end
                          
                          local hasKeinEinzug = hasToken (lineStripped, "\\keinEinzug", 1, true)
                          if inZentriertemBereich or hasKeinEinzug or erzwingeKeinEinzug then
                              if posBeitext == 0 then
                                  erzwingeKeinEinzug = false
                              end
                              
                              if hasKeinEinzug then
                                  lineOut = ltrim (lineOut:gsub ("%\\keinEinzug",""))
                              end
                              if inEingeruecktemBereich then
                                  rtfDocument:write("\\s24\\sl276\\slmult1\\ql\\widctlpar\\hyphpar0\\li397\\ri397\\lin397\\rin397\\fi0"..BlocksatzCode.."\\sb283\\sa0\\contextualspace\\ltrpar\\cf1"..kursiverBereich.."\\kerning1\\dbch\\af4\\langfe1031"..eventuellNeueSeiteCode.."\\dbch\\af8\\afs24\\loch\\f4\\fs24\\lang1031")
                              else
                                  rtfDocument:write("\\s16\\sl276\\slmult1\\ql\\widctlpar\\hyphpar0"..BlocksatzCode.."\\sb0\\sa0\\ltrpar\\cf1"..kursiverBereich.."\\kerning1\\dbch\\af6\\langfe1031"..eventuellNeueSeiteCode.."\\dbch\\af6\\afs24\\alang1031\\loch\\f4\\fs24\\sl276\\slmult1\\lang1031")
                              end
                          else
                              if inEingeruecktemBereich then
                                  rtfDocument:write("\\s25\\sl276\\slmult1\\ql\\widctlpar\\hyphpar0\\li397\\ri397\\lin397\\rin397\\fi397"..BlocksatzCode.."\\sb0\\sa283\\contextualspace\\ltrpar\\cf1"..kursiverBereich.."\\kerning1\\dbch\\af4\\langfe1031"..eventuellNeueSeiteCode.."\\dbch\\af8\\afs24\\loch\\f4\\fs24\\lang1031")
                              else
                                  rtfDocument:write("\\s22\\sl276\\slmult1\\ql\\widctlpar\\hyphpar0\\li0\\ri0\\lin0\\rin0\\fi397"..BlocksatzCode.."\\sb0\\sa0\\ltrpar\\cf1"..kursiverBereich.."\\kerning1\\dbch\\af6\\langfe1031"..eventuellNeueSeiteCode.."\\dbch\\af6\\afs24\\alang1031\\loch\\f4\\fs24\\sl276\\slmult1\\li0\\ri0\\lin0\\rin0\\fi397\\lang1031")
                              end
                          end
                          
                          if inZentriertemBereich then
                              rtfDocument:write("\\qc")
                          end
                      end
                      rtfDocument:write("{")
                      rtfDocument:write("\\dbch\\af6\\rtlch \\ltrch\\loch\n")
                  end
                  
                  local hasVertikalDehnen = hasToken (lineStripped, "\\vertikalDehnen", 1, true)
                  local hasLeerzeile      = hasToken (lineStripped, "\\Leerzeile", 1, true)
                  
                  if iLine == 1 then
                      kommentarEinbetten (rtfDocument, SPBuchsatzVersion,
                                          "\\line{\\b Bitte beachte:}\\line\\line Der RTF-Export von SPBuchsatz hat die meist-verwendeten Formatierungs-Codes automatisch umgewandelt.}\\par{\\cf0\\kerning0\\langfe1023\\alang1023\\dbch\\af9\\dbch\\af9\\loch\\f3\\lang1023}\\par{\\alang1023\\dbch\\af9\\charscalex100\\b0\\strike0\\i0\\outl0\\expnd0\\expndtw0\\kerning1\\shad0\\scaps0\\caps0\\cf1\\accnone\\dbch\\af4\\langfe1031\\ulnone\\ulc0\\lang1033\\fs20\\loch\\f3 SUCHE in diesem Text nach Schl\\uc2 \\u252\\'c3\\'bcsselworten, die mit Backslash \\\\ beginnen, und formatiere solche Stellen ggfs. manuell.\\uc1 }\\par{\\cf0\\kerning0\\langfe1023\\alang1023\\dbch\\af9\\dbch\\af9\\loch\\f3\\lang1023}\\par{\\alang1023\\dbch\\af9\\shad0\\ulnone\\ulc0\\langfe1031\\dbch\\af4\\accnone\\cf1\\scaps0\\caps0\\charscalex100\\kerning1\\expnd0\\expndtw0\\outl0\\i0\\strike0\\b0\\loch\\f3\\fs20\\lang1033 Beispielsweise werden \\\\kursiv\\{\\}-Bereiche und der Befehl \\\\Lombarde\\{\\} umgewandelt, aber {\\i keiner} der Schriftgr\\u246\\'3\\u223\\'3e-Befehle, also {\\i nicht} dieser Befehl:\\line \\{\\\\lombardengross \\u8230\\'3f\\}\\uc1 } \\line \\par{\\alang1023\\dbch\\af9\\shad0\\ulnone\\ulc0\\langfe1031\\dbch\\af4\\accnone\\cf1\\scaps0\\caps0\\charscalex100\\kerning1\\expnd0\\expndtw0\\outl0\\i0\\strike0\\b0\\loch\\f3\\fs20\\lang1033 Suche AUCH nach evtl. im Text auftretenden, geschweiften Klammern: \\{ oder \\}.\\uc1 } \\par{\\alang1023\\dbch\\af9\\shad0\\ulnone\\ulc0\\langfe1031\\dbch\\af4\\accnone\\cf1\\scaps0\\caps0\\charscalex100\\kerning1\\expnd0\\expndtw0\\outl0\\i0\\strike0\\b0\\loch\\f3\\fs20\\lang1033 In seltenen F\\uc2 \\u228\'3llen gelangen solche \\uc2 \\u252\\'c3\\'bcberfl\\uc2 \\u252\\'c3\\'bcssigen Zeichen in die RTF-Datei.\\line")
                  end

                  
                  
                  if (not hasLeerzeile) and (not hasVertikalDehnen) then
                  
                  
                      -- Folgende codiereSchalter()-Aufrufe stellen sicher, dass vor dem write()
                      -- die Befehle \kursiv bzw. \Normalschrift und \Lombarde exportiert werden:

                      if inKursivemBereich then
                          lineOut = codiereSchalter (lineOut, "\\Normalschrift", "\\i0",   true, false, false, 0, true, false)
                          lineOut = codiereSchalter (lineOut, "\\kursiv",        "\\i",    true, false, false, 0, true, true)  -- kursiv kann in Normalschrift eingekapselt sein.
                      else
                          lineOut = codiereSchalter (lineOut, "\\kursiv",        "\\i",    true, false, false, 0, true, false)
                          lineOut = codiereSchalter (lineOut, "\\Normalschrift", "\\i0",   true, false, false, 0, true, true)  -- Normalschrift kann in kursiv eingekapselt sein.
                      end
                      
                      lineOut     = codiereSchalter (lineOut, "\\Lombarde",      "\\cs12", true, false, false, 0, true, true)
                      
                      
                      -- Gegebenenfalls die extra Buchteil-Zeile voranstellen:

                      if extraBuchteilZeileVorab then
                          lineOut = codiereSchalter (extraBuchteilZeileVorab, "\\kursiv", "\\i", true, false, false, 0, true, true)
                              .."\\line \\line \\line "
                              .."}{\\cf1\\b0\\kerning1\\dbch\\af4\\langfe1031\\afs42\\rtlch \\ltrch\\loch\\fs42\\lang1031\\loch\\f4\\hich\\af4 \n"
                              ..lineOut
                      end

                      
                      -- Folgender Aufruf stellt sicher, dass vor dem write()
                      -- (1.) der Befehle \fett exportiert wird und
                      -- (2.) die lineOut ordnungsgemäß escaped wird,
                      --      wobei diese Codes (und ihre zugehörigen }-Klammern) nicht escaped werden: {\i, {\i0, {\cs12

                      lineOut = codiereSchalter (lineOut, "\\fett", "\\b", false, false, false, 0, true, true)

                      
                      -- Nun wird (ohne Escaping) der Code für \Lombarde vervollständigt:

                      lineOut = lineOut:gsub ("\\cs12 ", "\\cs12"..LombardenCode.."\n")
                                       :gsub ("\\u92\\'3fAbsatzKeinEinzug", "\\par\n")
                      

                      
                      if hasKapitelFortlaufend then
                          rtfDocument:write ("\\line")
                      end
                      
                      
                      rtfDocument:write (lineOut)

                      
                      if hasKapitelFortlaufend then
                          rtfDocument:write ("\\line")
                      end
                  end
                  
                  if posBeitext > 0 then
                      rtfDocument:write ("\\line\\line")
                      inZentriertemBereich = false
                      inKursivemBereich    = false
                  end
              
                  if hasVertikalDehnen then
                      kommentarEinbetten (rtfDocument, SPBuchsatzVersion,
                          "In der Print-Ausgabe ist nachfolgender Text am Fu\\uc2 \\u223\\'c3\\'9f der Seite gesetzt.")
                  end
                  
                  if imageWanted or
                     hasToken (lineStripped, "\\BildHoehe", 1, true) or
                     hasToken (lineStripped, "\\BildBreite", 1, true)
                  then
                      kommentarEinbetten (rtfDocument, SPBuchsatzVersion,
                          "In der Print-Ausgabe ist hier ein Bild im Text.")
                  end
                  
                  if besondereBildSeite then
                      kommentarEinbetten (rtfDocument, SPBuchsatzVersion,
                          "In der Print-Ausgabe hast du hier ein formatf\\uc2 \\u252\\'c3\\'bcllendes Bild.")
                      neueSeiteAngefordert = true
                  end
                  
                  rtfDocument:write(absatzAbschlussCodes)
              end
              
          end
          
          ::endOfLineLoop::
      end
      
      kommentarEinbetten (rtfDocument, SPBuchsatzVersion,
                          "\\line{\\b Bitte beachte:}\\line\\line Falls Du Inhaltshinweise / Content Notes in Deinem Buch hast, so trage sie bitte auch in diese Datei ein.\\line\\line Der RTF-Export kann sie aktuell noch nicht exportieren, daher fehlen sie gegebenenfalls in dieser Datei.\\line")
      
      rtfDocument:write("}\n")
      rtfDocument:close()
      rawFile:close()
      
      print ("]\n")
  end
\end{luacode*}
