% Datei: SPBuchsatz_System/Helferlein/SPBuchsatz_auxiliary.lua
%
%  Diese Datei enthält interne Unterprogrammroutinen und Datenstrukturen,
%  die von SPBuchsatz-Funktionen aufgerufen werden.
%
%
%  CAVEAT: File providing SPBuchsatz system code.


\begin{luacode*}

  local lfs = require "lfs"

  function strSimplified (strIn)
    local strOut = string.strip (strIn:gsub("%s+", " "))
    return strOut
  end

  -- the following auxiliary functions return Zero when nothing found (not nil).
  function returnValueOrZero (pos)
    if (pos) then
      return pos
    end
    return 0
  end
  
  
  function findPatternFrom (str, pattern, pos)
    if (pos == nil) or (pos == 0) or (str == nil) or (pattern == nil) then
      return 0
    end
    return returnValueOrZero (str:find (pattern, pos, false))
  end
  
  
  function findPatternFirst (str, pattern)
    return findPatternFrom (str, pattern, 1)
  end
  
  
  function hasPattern (str, pattern)
    local pos = findPatternFirst (str, pattern)
    return (pos > 0)
  end
  
  
  function startsWithPattern (line, token)
    local pos = findPatternFirst (line, token)
    return (pos == 1)
  end
  
  
  function findFrom (str, token, pos)
    if (pos == nil) or (pos == 0) or (str == nil) or (token == nil) then
      return 0
    end
    return returnValueOrZero (str:find (token, pos, true))
  end
  
  
  function findFirst (str, token)
    return findFrom (str, token, 1)
  end
  
  
  function hasToken (str, token)
    local pos = findFirst (str, token)
    return (pos > 0)
  end
  
  
  -- Note: findAfter searches from oldPos+1 (but not from oldPos)
  function findAfter (str, token, oldPos)
    if (oldPos == nil) or (oldPos == 0) or (str == nil) or (token == nil) then
      return 0
    end
    return findFrom (str, token, oldPos+1)
  end
  
  
  function startsWith (line, token)
    local pos = findFirst (line, token)
    return (pos == 1)
  end
  
  
  function endsWith (line, token)
    local lengthOfLine  = line:len ()
    local lengthOfToken = token:len ()
    if lengthOfLine < lengthOfToken then
      return false
    end
    if lengthOfLine == lengthOfToken then
      return (line == token)
    end
    return (line:sub (lengthOfLine - lengthOfToken + 1) == token)
  end
  
  
  function findLast (str, token)
    local pos = findFirst (str, token)
    local nextPos = pos
    while nextPos > 0 do
      pos = nextPos
      nextPos = findAfter (str, token, pos)
    end
    return pos
  end
  
  
  function findSecondLast (str, token)
    local pos = findFirst (str, token)
    if pos == 0 then
      return 0
    end
    pos = 0
    local prevPos = 0
    local nextPos = findAfter (str, token, pos)
    while nextPos > 0 do
      if prevPos > 0 then
        pos = prevPos
      end
      prevPos = nextPos
      nextPos = findAfter (str, token, nextPos)
    end
    return pos
  end

  
  function findFirstComment (line)
    local pos = findFirst ('%')
    if pos > 0 then
      if line:sub (pos-1, pos-1) == "\\" then
        return 0
      end
    end
    return pos
  end
  
  
  function splitString (s, sep)
    return coroutine.wrap (function ()
      if (s == '') or (sep == '') then
        coroutine.yield (s)
        return
      end
      local lasti = 1
      for v,i in s:gmatch ('(.-)'..sep..'()') do
        coroutine.yield (v)
        lasti = i
      end
      coroutine.yield (s:sub (lasti))
    end)
  end
  

  -- Search (from startIdx) for a closing brace, skip any extra pairs of opening/closing braces found.
  -- 
  -- Returns idx of closing brace, or Zero, if matching closing brace not found.
  -- 
  function findClosingBraceAndRemainingDepthExt (line, startIdx, openingBrace, closingBrace)

    local lineLength = line:len ()
    
    local depth = 1

    local idx = startIdx
    repeat
      local posOpen  = findFrom (line, openingBrace, idx)
      local posClose = findFrom (line, closingBrace, idx)
      
      if posClose > 0 then
        if (posOpen == 0) or (posClose < posOpen) then
          if depth == 1 then
            return posClose, 0
          end
          depth = depth -1
          idx   = posClose
        else
          depth = depth +1
          idx   = posOpen
        end
      else
        if posOpen > 0 then
          depth = depth +1
          idx   = posOpen
        end
      end
      
      idx = idx+1
    until (idx > lineLength)
    
    return 0, depth
  end

  
  function findClosingBraceAndRemainingDepth (line)
    return findClosingBraceAndRemainingDepthExt (line, 1, "{", "}")
  end



  function findClosingBraceExt (line, startIdx, openingBrace, closingBrace)
    local pos, iDummy = findClosingBraceAndRemainingDepthExt (line, startIdx, openingBrace, closingBrace)
    return pos
  end

  function findClosingBrace (line)
    return findClosingBraceExt (line, 1, "{", "}")
  end
  
  

  spInternUmlauts = "ÄÖÜÁÉÍÓÚÀÈÌÒÙÂÊÎÔÛ"
  
  function isUmlaut (c)
    return hasToken (spInternUmlauts, c)
  end
  
  function isDigitOrDot (c)
    return ((c >= "0") and (c <= "9")) or (c == ".")
  end



  -- Internal code for Table handling - - - - - - - - - - - - - - - - - - -
  
  function isVLineToken (c)
    return ((c == "|") or (c == "i") or (c == "I") or (c == "A") or (c == "B") or (c == "C") or (c == "D") or (c == "E") or (c == "F"))
  end
  
  function isColumnAlignToken (c)
    return ((c == "l") or (c == "b") or (c == "r") or (c == "z") or (c == "c"))
  end
  

  -- internal variables for calculating and caching multicolumn widths
  spInternColumnWidthCommands   = {}
  spInternMultiColumnWidthCount = 0
  spInternKnownColumnWidths     = {}


  -- internal storage of column specification of current table
  spInternCurrentColumnSpec = 
  {
    -- Five methods used at runtime to retrieve information:
    isEmpty     = function (self)
                    return (#self.data == 0)
                  end,
    
    getCount    = function (self)
                    return #self.data
                  end,

    getLeft     = function (self, idx)
                    if self:idxIsFine (idx) then
                      return self.data [idx].leftValue
                    end
                    return ""
                  end,
    
    getRight    = function (self, idx)
                    if self:idxIsFine (idx) then
                      return self.data [idx].rightValue
                    end
                    return ""
                  end,
    
    getAlign    = function (self, idx)
                    if self:idxIsFine (idx) then
                      return self.data [idx].alignValue
                    end
                    return ""
                  end,

    -- Methods used during initialization of the table:
    clearData   = function (self)
                    self.data = {}
                  end,
    
    addColumn   = function (self, leftLine, columnAlign, rightLine)
                    self.data [self:getCount () + 1] = {leftValue  = leftLine, 
                                                        alignValue = columnAlign, 
                                                        rightValue = rightLine}
                    --print ("specifiers added: ", leftLine, columnAlign, rightLine)
                  end,
  
    getCurLeft  = function (self)
                    return self:getLeft (self:getCount ())
                  end,
    
    getCurRight = function (self)
                    return self:getRight (self:getCount ())
                  end,
    
    getCurAlign = function (self)
                    return self:getAlign (self:getCount ())
                  end,

    setCurLeft  = function (self, value)
                    if self:getCount () > 0 then
                      self.data [self:getCount ()].leftValue = value
                    end
                  end,
    
    setCurRight = function (self, value)
                    if self:getCount () > 0 then
                      self.data [self:getCount ()].rightValue = value
                    end
                  end,
    
    setCurAlign = function (self, value)
                    if self:getCount () > 0 then
                      self.data [self:getCount ()].alignValue = value
                    end
                  end,

    -- private methods and internal data:
    idxIsFine   = function (self, idx)
                    return (idx > 0) and (idx <= self:getCount ())
                  end,
    data = {}
  }
  
  
  function isVLineTokenOrDash (c)
    return (isVLineToken (c) or (c == "-"))
  end
  
  
  function startsWithHLineCommand (line)
    return startsWith (line, "\\Linie") or startsWith (line, "\\hline")
  end
  
  
  function startsWithVSpaceCommand (line)
    return startsWith (line, "\\Abstand")
  end
  
  
  function startsWithBeginParTableCommand (line)
    return startsWith (line, "\\BeginnTabelle")
  end
  function startsWithAnyBeginTableCommand (line)
    return startsWithBeginParTableCommand (line) or startsWith (line, "\\BeginnKleintabelle")
  end
  function startsWithEndParTableCommand (line)
    return startsWith (line, "\\EndeTabelle")
  end
  function startsWithAnyEndTableCommand (line)
    return startsWithEndParTableCommand (line) or startsWith (line, "\\EndeKleintabelle")
  end
  
  
  spInternStartOfCellFormatCommand  = "\\Zellenformat{"  -- the { is included intentionally
  spInternStartOfMulticolumnCommand = "\\mehrspaltig{"   -- the { is included intentionally
  spInternStartOfColumnWidthCommand = "\\Spaltenbreite{" -- the { is included intentionally
  spInternStartOfAllcolumnCommand   = "\\allspaltig"     -- the { is omitted intentionally   :)
      
    
  function getPostfixForCalculatedColumnWidth (iColumn, columnCountParameter)
    local postfix = ""
    -- check if we have cached a matching multicolumn width:
    local tmpKnownKey = tostring (iColumn)..":"..columnCountParameter
    
    if spInternKnownColumnWidths [tmpKnownKey] ~= nil then
      postfix = spInternKnownColumnWidths [tmpKnownKey]
    else
      -- This calculation needs spInternMultiColumnWidthCount to be Zero-based
      -- so we increment it /after/ composing the postfix:
      if spInternMultiColumnWidthCount < 26 then
        postfix = string.char (65 + spInternMultiColumnWidthCount)
      else
        local lastDigit = math.floor(spInternMultiColumnWidthCount % 26)
        postfix = string.char (64 + math.floor(spInternMultiColumnWidthCount / 26)) .. string.char (65 + lastDigit)
      end
      spInternColumnWidthCommands[#spInternColumnWidthCommands + 1] = "\\errechneBreiteTMx{"..postfix.."}{"..iColumn.."}{"..columnCountParameter.."}"
      spInternKnownColumnWidths [tmpKnownKey] = postfix
      spInternMultiColumnWidthCount = spInternMultiColumnWidthCount + 1
    end
    return postfix
  end
  
  function getPostfixForCalculatedColumnWidthPlusGap (isLeft, iColumn)
    local postfix = ""
    -- check if we have cached a matching multicolumn width:
    
    local lineTokenLeft  = ""
    local lineTokenRight = ""
    local tmpKnownKey = tostring (iColumn)..":"

    if isLeft then
      tmpKnownKey = tmpKnownKey.."LEFT"
    else
      tmpKnownKey = tmpKnownKey.."RIGHT"
    end
    
    if isLeft then
      lineTokenRight = spInternCurrentColumnSpec:getRight (iColumn)
    elseif iColumn == 1 then
      lineTokenLeft = spInternCurrentColumnSpec:getLeft (iColumn)
    end
    
    if spInternKnownColumnWidths [tmpKnownKey] ~= nil then
      postfix = spInternKnownColumnWidths [tmpKnownKey]
    else
      -- This calculation needs spInternMultiColumnWidthCount to be Zero-based
      -- so we increment it /after/ composing the postfix:
      if spInternMultiColumnWidthCount < 26 then
        postfix = string.char (65 + spInternMultiColumnWidthCount)
      else
        local lastDigit = math.floor(spInternMultiColumnWidthCount % 26)
        postfix = string.char (64 + math.floor(spInternMultiColumnWidthCount / 26)) .. string.char (65 + lastDigit)
      end
      spInternColumnWidthCommands[#spInternColumnWidthCommands + 1] = "\\errechneBreiteTZx{"..postfix.."}{"..iColumn.."}{"..lineTokenLeft.."}{"..lineTokenRight.."}"
      spInternKnownColumnWidths [tmpKnownKey] = postfix
      spInternMultiColumnWidthCount = spInternMultiColumnWidthCount + 1
    end
    return postfix
  end
    
  
  function preprocessTableLine (line, columnCount, weAreInParTable)
    local lineNew  = ""
    local iColumn  = 1
    local lastChar = ""

    -- Traverse every char in the line to process (single or multiple)
    -- occurences of multicolumn commands:
    local idx = 1
    while idx <= #line do
      
      local startWithCellFormatCommand = startsWith (line:sub (idx), spInternStartOfCellFormatCommand)
      local startWithAllcolumnCommand = startsWith (line:sub (idx), spInternStartOfAllcolumnCommand)
      
      if startWithCellFormatCommand or startWithAllcolumnCommand or startsWith (line:sub (idx), spInternStartOfMulticolumnCommand) then
        lineNew = lineNew.."\\multicolumn{"
          
        local columnCountParameter = ""

        if startWithCellFormatCommand then
          columnCountParameter = "1"
          lineNew = lineNew..columnCountParameter.."}"
          -- We insert "{ … }" into the line, because the cell format command
          -- does not have the cells text as parameter, but the multicolumn
          -- command needs that.
          local posOpen = findFrom (line, "{", idx)
          if posOpen > 0 then  
            local posClose = findFrom (line, "}", posOpen)
            if posClose > 0 then
              local tmpIdx = posClose
              local posEnd = 0
              while (posEnd == 0) and (tmpIdx <= line:len ()) do
                posEnd = findFrom (line, "&", tmpIdx)
                if posEnd > 0 then
                  if (line:sub (posEnd-1, posEnd-1) == "\\") then
                    posEnd = 0
                    tmpIdx = posEnd + 1
                  end
                else
                  tmpIdx = line:len () + 1
                end
              end
              if posEnd == 0 then
                posEnd = findFrom (line, "\\\\", idx)
              end
              if posEnd > 0 then
                line = line:sub (1, posClose) .. "{" .. line:sub (posClose+1, posEnd-1) .. "}" .. line:sub (posEnd)
                idx  = posOpen
              end
            end
          end
        elseif startWithAllcolumnCommand then
          columnCountParameter = tostring (columnCount)
          lineNew = lineNew..columnCountParameter.."}"
          idx = idx + spInternStartOfAllcolumnCommand:len ()
        else
          idx = idx + spInternStartOfMulticolumnCommand:len ()
          
          -- copy and save 1st parameter
          while idx <= #line and not (line:sub (idx, idx) == "}") do
            local c = line:sub (idx, idx)
            lineNew              = lineNew..c
            columnCountParameter = columnCountParameter..c
            idx = idx + 1
          end
        end
        
        while idx <= #line and not (line:sub (idx, idx) == "{") do
          lineNew = lineNew..line:sub (idx, idx)
          idx = idx + 1
        end
        
        lineNew = lineNew.."{"
        idx = idx + 1
        
        local postfix = ""

        if weAreInParTable then
          postfix = getPostfixForCalculatedColumnWidth (iColumn, columnCountParameter)
        end
        
        local columnsInMultitable = tonumber (columnCountParameter)

        if columnsInMultitable == nil then
          columnsInMultitable = 1
        end
        
        local alignParameter = ""
        
        
        while idx <= #line and not (line:sub (idx, idx) == "}") do
          
          local c = line:sub (idx, idx)
          
          if isVLineTokenOrDash (c) then
            
            alignParameter = alignParameter..c
            
          elseif isColumnAlignToken (c) then
            
            -- the very first column can use left and right VLines, the next ones just a right VLine:
            if (iColumn == 1) and not isVLineTokenOrDash (alignParameter:sub (alignParameter:len (), alignParameter:len ())) then
              alignParameter = alignParameter..spInternCurrentColumnSpec:getLeft (iColumn)
            end
            
            if weAreInParTable then
              alignParameter = alignParameter..getLongTableColumnPrefix (c).."p{\\HilfsbreiteTMx"..postfix.." mm}"
            else
              alignParameter = alignParameter..c:gsub ("z", "c")
            end
            
            local followingChars = line:sub (idx + 1):gsub (" ", "")
            
            if not isVLineTokenOrDash (followingChars:sub (1, 1)) then
              alignParameter = alignParameter..spInternCurrentColumnSpec:getRight (iColumn + columnsInMultitable - 1)
            end
          end
          idx = idx + 1
        end
        
        -- Remove dashes, they were only there to prevent use of default VLines for this line:
        lineNew = lineNew .. alignParameter:gsub ("-", "") .. "}"
        
        idx = idx + 1
        
        iColumn = iColumn + columnsInMultitable - 1
      else
        local c = line:sub (idx, idx)
        lineNew = lineNew..c
        if (c == "&") and not (lastChar == "\\") then
          iColumn = iColumn + 1
        end
        lastChar = c
        idx = idx + 1
      end
    end
    
    if weAreInParTable then
      local posWidthOfColumn = findFirst (lineNew, spInternStartOfColumnWidthCommand)
      while (posWidthOfColumn > 0) do
        local posStart = posWidthOfColumn + spInternStartOfColumnWidthCommand:len ()
        local posClose = findFrom (lineNew, "}", posStart)
        if posClose > 0 then
          local columnIdx = tonumber (lineNew:sub (posStart, posClose-1))
          if columnIdx then
            postfix = getPostfixForCalculatedColumnWidth (columnIdx, 1)
            local lineTmp = lineNew:sub (1, posWidthOfColumn-1).."\\HilfsbreiteTMx"..postfix.." mm"..lineNew:sub (posClose+1)
            lineNew = lineTmp
            posWidthOfColumn = findFirst (lineNew, spInternStartOfColumnWidthCommand)
          end
        end
      end
    end
    
    return lineNew
  end


  function addPrefixPostfixToTableCell (line, lineStripped, cellPrefix, cellPostfix)
    local lineNew = ""

    if startsWith (lineStripped, "\\multicolumn{") then
      local posOpen = findFirst (line, "{")
      if posOpen == 0 then
        return line
      end
      local posClose = findFrom (line, "}", posOpen)
      if posClose == 0 then
        return line
      end
      posOpen = findFrom (line, "{", posClose)
      if posOpen == 0 then
        return line
      end
      posClose = findClosingBrace (line:sub (posOpen + 1))
      if posClose == 0 then
        return line
      end
      posClose = posOpen + posClose
      posOpen = findFrom (line, "{", posClose)
      if posOpen == 0 then
        return line
      end
      local trailingPart = line:sub (posOpen + 1)
      posClose = findClosingBrace (trailingPart)
      if posClose == 0 then
        return line
      end
      lineNew = line:sub (1, posOpen) .. cellPrefix .. trailingPart:sub (1, posClose-1) .. cellPostfix .. trailingPart:sub (posClose)
    else
      local posEOL = findFirst (line, "\\\\")
      if posEOL > 0 then
        lineNew = cellPrefix .. line:sub (1, posEOL-1):gsub ('^%s*(.-)%s*$', '%1') .. cellPostfix .. " " .. line:sub (posEOL)
      else
        lineNew = cellPrefix .. line .. cellPostfix
      end
    end
    
    return lineNew
  end


  function addPrefixPostfixToTableCells (line, cellPrefix, cellPostfix)
    
    local parts = splitString (line:gsub ("\\&", "spAmpersand"), "&")

    local lineNew = ""
    local idx = 0
    for partRaw in parts do
      local part = partRaw:gsub('^%s*(.-)%s*$', '%1')
      idx = idx + 1
      if idx > 1 then
        lineNew = lineNew.." & "
      end
      local partStripped = part:gsub (" ", "")
      if (startsWithHLineCommand (partStripped) or startsWithVSpaceCommand (partStripped)) then
        lineNew = lineNew .. part
      else
        lineNew = lineNew .. addPrefixPostfixToTableCell (part, partStripped, cellPrefix, cellPostfix)
      end
    end
    
    if idx == 0 then
      return addPrefixPostfixToTableCell (line, line:gsub (" ", ""), cellPrefix, cellPostfix)
    end
    
    return lineNew:gsub ("spAmpersand", "\\&")
  end
  
  
  function preprocessColumnSpecifiers (columnSpecifiers)

    spInternCurrentColumnSpec:clearData ()

--print(" - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - ")
--print(columnSpecifiers)
    
    local specifiers = ""
    local idx = 1
    while idx <= #columnSpecifiers do
      local remainingChars = columnSpecifiers:sub (idx)
--print("remaining: "..remainingChars)
      local c = remainingChars:sub (1, 1)
      if c == "z" then
        c = "c"
      end
      if isColumnAlignToken (c) then
        if spInternCurrentColumnSpec:isEmpty () or not (spInternCurrentColumnSpec:getCurAlign () == "") then
          spInternCurrentColumnSpec:addColumn ("", c, "")
        else
          spInternCurrentColumnSpec:setCurAlign (c)
        end
      elseif isDigitOrDot (c) then
        -- do nothing (We are not storing the widths in the table.)
      elseif startsWithPattern (remainingChars, "Breite%u") then
        -- do nothing (We are not storing the widths in the table.)
        idx = idx + 6
      else
        -- This is a line indicator, either a left border or a right one:
        if spInternCurrentColumnSpec:isEmpty () then
          spInternCurrentColumnSpec:addColumn (c, "", "")
        else
          if spInternCurrentColumnSpec:getCurRight () == "" then
            spInternCurrentColumnSpec:setCurRight (c)
          else
            spInternCurrentColumnSpec:addColumn ("", "", c)
          end
        end
      end
      specifiers = specifiers..c
      idx = idx + 1
    end
    --print ("    raw:    ", columnSpecifiers)
    --print ("    specs:  ", specifiers)
    --for idx = 1, spInternCurrentColumnSpec:getCount () do
    --  print ("    column: "..spInternCurrentColumnSpec:getLeft (idx).."    "..spInternCurrentColumnSpec:getAlign (idx).."    "..spInternCurrentColumnSpec:getRight (idx))
    --end
    return specifiers, spInternCurrentColumnSpec:getCount ()
  end

  
  function createVSpaceLine (verticalSpacing, columnCount)
    local postfixLeft  = getPostfixForCalculatedColumnWidthPlusGap (true,  1)
    local postfixRight = getPostfixForCalculatedColumnWidthPlusGap (false, columnCount)
    return "\\cmidrule["..verticalSpacing.."\\normalbaselineskip](l{0pt}r{\\HilfsbreiteTZx"..postfixLeft.." mm}){1-1}\\cmidrule["..verticalSpacing.."\\normalbaselineskip](r{0pt}l{\\HilfsbreiteTZx"..postfixRight.." mm}){"..columnCount.."-"..columnCount.."}"
  end
  

  function preprocessTableLines (rawLines)
    local newLines = {}
    
    newLines [1] = ""  -- reserved for columnWidthCommand, see end of function
    newLines [2] = ""  -- reserved for HLine indicator
    newLines [3] = ""  -- reserved for upper VSpace command before table starts

    local nextVerticalSpacing  = ""
    local previousLineStripped = ""
    local mustSkipNextAutomaticVSpacing = false
    
    local weAreInParTable = false
    local columnCount     = 0
    local tableHasHLines  = false
    
    -- Global variables, read and set by preprocessTableLine(), too:
    spInternColumnWidthCommands   = {}
    spInternMultiColumnWidthCount = 0
    spInternKnownColumnWidths     = {}
    
    
    for lineIdx = 1, #rawLines do 
      
      local spInternUseVSpacingRow = weAreInParTable  -- vertical spacing is done by inserting a 'magic' row into the table
      
      local line                    = rawLines [lineIdx]
      local lineStripped            = line:gsub (" ", "")

      if (lineStripped:len () == 0) or (findFirstComment (lineStripped) == 1) then
        goto endOfLoop -- Not store this line in previousLineStripped.
      end
      
      local thisIsBeginTableCommand = startsWithAnyBeginTableCommand (lineStripped)
      local thisIsEndTableCommand   = startsWithAnyEndTableCommand   (lineStripped)
      local thisIsEndFootCommand    = startsWith (lineStripped, "\\EndeLetzteFusszeilen") or startsWith (lineStripped, "\\EndeFusszeilen")
      local thisIsVSpaceCommand     = startsWithVSpaceCommand        (lineStripped)
      local thisIsHLineCommand      = startsWithHLineCommand         (lineStripped)

      local prevIsBeginTableCommand = startsWithAnyBeginTableCommand (previousLineStripped)
      local prevIsHLineCommand      = startsWithHLineCommand         (previousLineStripped)
      
      if thisIsBeginTableCommand then
        weAreInParTable = startsWithBeginParTableCommand (lineStripped)
      end

      local cellPrefix  = ""
      local cellPostfix = ""
      
      -- We are not using \addlinespace[…] here because \addlinespace and
      -- vertical lines do not not work together well.
      -- This is the safest way: Add two cmidrule's at the first / last cells of a line,
      -- the heights of the rules is the desired amount of vertical gap, their widths
      -- are minimal: just the respective width of vertical lines shown at the sides of the cells.
      -- (If no line exists the width is Zero.)
      
      if mustSkipNextAutomaticVSpacing then
        if spInternUseVSpacingRow then
          nextVerticalSpacing = ""
        end
        mustSkipNextAutomaticVSpacing = false
      elseif (nextVerticalSpacing == "") and 
             ((thisIsHLineCommand and not prevIsBeginTableCommand) or (prevIsHLineCommand and not (thisIsEndTableCommand or thisIsEndFootCommand))) then
        nextVerticalSpacing = "\\spAbstandInTabelle"
      end

      
      if thisIsBeginTableCommand then

        -- find out how many columns we have:
        local posOpen = findFirst (lineStripped, "{")
        if posOpen > 0 then
          local trailingPart = lineStripped:sub (posOpen + 1)
          local posClose = findClosingBrace (trailingPart)
          if posClose > 1 then
            local columnSpecifiers = trailingPart:sub (1, posClose -1)
            -- Actually we do not care about the adjusting of c to z here,
            -- we just need to know the columnCount:
            columnSpecifiers, columnCount = preprocessColumnSpecifiers (columnSpecifiers)
          end
        end
        
        -- store the line:
        newLines [#newLines + 1] = line
        
      elseif thisIsVSpaceCommand then
        
        mustSkipNextAutomaticVSpacing = true      
        
        local appendText = ""
        if startsWith (lineStripped, "\\AbstandAus") then
          nextVerticalSpacing  = ""
        elseif startsWith (lineStripped, "\\Abstand{") then
          
          local verticalSpacing = nextVerticalSpacing
          nextVerticalSpacing   = ""
          local trailingPart = lineStripped:sub (10)
          local posClose = findClosingBrace (trailingPart)
          
          if posClose > 1 then
            local value = trailingPart:sub (1, posClose-1):gsub(" ", "")
            
            if prevIsBeginTableCommand then
              -- add vertical spacing before the table:
              newLines [3] = "\\FaktorLeerzeile{"..value.."}"
              verticalSpacing = ""
            else
              verticalSpacing = value
              appendText = trailingPart:sub (posClose+1)
            end
          else
            verticalSpacing = ""
          end
          
          if spInternUseVSpacingRow then
            newLines [#newLines + 1] = createVSpaceLine (verticalSpacing, columnCount)..appendText
          else
            nextVerticalSpacing = verticalSpacing
          end
        end

        goto endOfLoop -- Not store this line in previousLineStripped.
        
      elseif thisIsHLineCommand then
        
        if not (nextVerticalSpacing == "") then
          
          if spInternUseVSpacingRow then
            newLines [#newLines + 1] = createVSpaceLine (nextVerticalSpacing, columnCount)
          else
            if prevIsHLineCommand then
              
              -- If we are NOT in VSpacingRow mode we ignore this spacing if also the previous line was a HLine.
              
            else
              -- We cannot set vertical spacing to a HLine so we now set it BELOW
              -- each of the cells of the PREVIOUS line: in newLines [#newLines]
              if weAreInParTable then
                cellPostfix = "\\ZellenabstandUnten{"..nextVerticalSpacing.."}"
              else
                -- This goes into the prefix, intentionally:
                cellPrefix = "\\setlength{\\spInternVSpacingInTable}{"..nextVerticalSpacing.."\\normalbaselineskip}\\addtolength{\\spInternVSpacingInTable}{0.3\\normalbaselineskip}\\rule[-\\spInternVSpacingInTable]{0pt}{\\spInternVSpacingInTable}"
              end
              newLines [#newLines] = addPrefixPostfixToTableCells (newLines [#newLines], cellPrefix, cellPostfix)
            end
          end
          nextVerticalSpacing = ""
        end

        newLines [#newLines + 1] = line
        tableHasHLines = true
        
      elseif thisIsEndTableCommand then
        
        newLines [#newLines + 1] = line
        if not (nextVerticalSpacing == "") then
          newLines [#newLines + 1] = "\\FaktorLeerzeile{"..nextVerticalSpacing.."}"
          nextVerticalSpacing = ""
        end
        
      else -- NORMAL TEXT LINE - - - - - - - - - - - - - - - - - -

        local lineNew = preprocessTableLine (line, columnCount, weAreInParTable)
        
        if nextVerticalSpacing == "" then
          newLines [#newLines + 1] = lineNew
        else
          if spInternUseVSpacingRow then
            newLines [#newLines + 1] = createVSpaceLine (nextVerticalSpacing, columnCount)
            newLines [#newLines + 1] = lineNew
          else
            if weAreInParTable then
              cellPrefix = "\\ZellenabstandOben{"..nextVerticalSpacing.."}"
            else
              cellPrefix = "\\setlength{\\spInternVSpacingInTable}{"..nextVerticalSpacing.."\\normalbaselineskip}\\addtolength{\\spInternVSpacingInTable}{0.7\\normalbaselineskip}\\rule{0pt}{\\spInternVSpacingInTable}"
            end
            newLines [#newLines + 1] = addPrefixPostfixToTableCells (lineNew, cellPrefix, cellPostfix)
          end
          nextVerticalSpacing = ""
        end
        
      end
      previousLineStripped = lineStripped
    
      ::endOfLoop::
    end -- for lineIdx
    

    if #spInternColumnWidthCommands > 0 then
      local columnWidthCommand = "\\def\\errechneHilfsbreiten{"
      for commandsIdx = 1, #spInternColumnWidthCommands do 
        columnWidthCommand = columnWidthCommand..spInternColumnWidthCommands [commandsIdx]
      end
      columnWidthCommand = columnWidthCommand.."}"
      newLines [1] = columnWidthCommand
    end

    if weAreInParTable and tableHasHLines then
      newLines [2] = "\\setzeAbsatztabelleMitHorizontalenLinien"
    end
      
    return newLines
  end
\end{luacode*}
  
  
  
% - Internal code for user command handling - - - - - - - - - - - - - - - -

\makeatletter
{\catcode`[=1 \catcode`]=2
  \catcode`#=11
\gdef\charHash[#]
\gdef\charPercent[\@percentchar]
]
\makeatother

\def\charBS{\\csname @backslashchar\\endcsname}

\begin{luacode*}
  
  function prepareForDebugOutput (line)
    local newLine   = ""
    local needSpace = false
    for idx = 1, #line do
      local origChar = line:sub (idx, idx)
      if origChar == "{" then
        newLine = newLine.."\\csname @charlb\\endcsname"
        needSpace = true
      elseif origChar == "}" then
        newLine = newLine.."\\csname @charrb\\endcsname"
        needSpace = true
        --elseif origChar == "\\" then
        --  newLine = newLine.."\\ " -- "\\charBS"
        needSpace = true
      elseif origChar == "#" then
        newLine = newLine.."\\charHash"
        needSpace = true
      elseif origChar == " " then
        newLine = newLine.."\\space "
        needSpace = false
      elseif origChar == "%" then
        newLine = newLine.."\\charPercent "  -- "\\% "
          needSpace = false
      else
        if needSpace then
          newLine = newLine.." "
          needSpace = false
        end
        newLine = newLine..origChar
      end
    end
    return newLine
  end
  
  
  local lBrace      = "\\csname @charlb\\endcsname "
  local rBrace      = "\\csname @charrb\\endcsname "
  local charBS      = "\\ " --"\\charBS "
  local charHash    = "\\charHash "
  
  local longIndent  = "\\space\\space\\space\\space\\space\\space\\space\\space\\space\\space\\space\\space\\space\\space\\space\\space"
  local shortIndent = "\\space\\space\\space\\space"
  
  function haltWithError (lineNumber, fileName, line, shortText, longText)
    tex.print("\\PackageError{SPBuchsatz}"..
              "{"..
             shortText.."."..
             "\\MessageBreak\\space"..
             "\\MessageBreak Der Fehler trat auf beim Verarbeiten der Datei:"..
             "\\MessageBreak\\space"..
             "\\MessageBreak "..fileName..
             "\\MessageBreak\\space"..
             "\\MessageBreak Ungefähre Zeilennummer des Fehlers: "..tostring (lineNumber)..
             "\\MessageBreak\\space"..
             "\\MessageBreak Wahrscheinlich fehlerhafte Zeile:"..
             "\\MessageBreak\\space"..
             "\\MessageBreak "..prepareForDebugOutput (line)..
             "\\MessageBreak\\space"..
             "\\MessageBreak\\space"..
             "\\MessageBreak Bitte beende den Textsatz mit X und Eingabetaste"..
             "}{"..
             "Bitte die X-Taste, dann die Eingabetaste."..
             "}"
             )
  end
  
  
  local spStartTokenOfUserCommandDefinition = "\\meinBefehl"
  local spEndTokenOfUserCommandDefinition   = "\\EndeBefehl"
  
  spInternUserCommands = {}
  
  
  -- Parameter starts by a command name, and might have more chars trailing which can be ignored
  -- 
  -- This function assumes the tables is already sorted by the lengths of the names, longest first.
  -- 
  -- Returns the commannd table, or nil.
  -- 
  function findBestMatchingUserCommand (lineNumber, fileName, commandNameAndMore)
    for i,v in ipairs(spInternUserCommands) do
      if startsWith (commandNameAndMore, v.name) then
        return v
      end
    end
    haltWithError (lineNumber, fileName, "Befehl"..commandNameAndMore,
                   "Dieser Befehl existiert nicht",
                   "Der Befehl \\meinBefehl muss zu Beginn einer Zeile aufgerufen werden"..
                  "\\MessageBreak und zwischen \\\\space und meinBefehl darf keine Lücke sein."
                  )
    return nil
  end
  
  
  -- Looks if the exact commandName was stored already,
  -- returns true or false.
  -- 
  function findUserCommand (commandName)
    local iCommand = 0
    for i,v in ipairs(spInternUserCommands) do
      iCommand = iCommand + 1
      if v.name == commandName then
        return iCommand
      end
    end
    return 0
  end
  
  
  function storeUserCommand (commandName, paramCount, contents)
    line = "Registriere Befehl"
    local iCommand = findUserCommand (commandName) 
    if iCommand == 0 then
      iCommand = #spInternUserCommands+1
    else
      line = line.." erneut"
    end
    line = line..": \\Befehl"..commandName
    if paramCount > 0 then
      for idx = 1, paramCount do
        line = line.."{…}"
      end
    end
    print (line)
    local command = {}
    command ["name"] = commandName
    command ["paramCount"] = paramCount
    command ["contents"] = contents
    spInternUserCommands [iCommand] = command
  end

  
  spCurrentlyInDefinitionOfUserCommand   = false
  spCurrentlyParsedUserCommandName       = ""
  spCurrentlyParsedUserCommandParamCount = 0
  spRawLinesOfUserCommand                = {}

  local generalSyntaxErrorForUserCommandsShort = "Nach \\meinBefehl fehlt […]"..lBrace.."…"..rBrace..lBrace.."\\space oder "..lBrace.."…"..rBrace
  local generalSyntaxErrorForUserCommandsLong  = "Unmittelbar nach der Anweisung \\meinBefehl müssen entweder zuerst zwei eckige"..
                                                 "\\MessageBreak Klammern mit Angabe der von dir erwarteten Parameterzahl folgen, z.B. [2],"..
                                                 "\\MessageBreak danach folgt dann der Name in geschweiften Klammern "..lBrace..rBrace..","..
                                                 "\\MessageBreak oder der Name folgt sofort nach \\meinBefehl."



  -- Parse the line and handle any user command definitions found.
  --
  -- Returns:
  --   1.  false, if error occured, true if no error
  --   2.  true, if (part of) a user command definition was found, false if none found
  -- 
  function parseLineAndStoreUserCommandDefinitionIfAny (lineNumber, fileName, line, lineStripped)
    if startsWith (lineStripped, spStartTokenOfUserCommandDefinition) then
      
      spRawLinesOfUserCommand = {}
      
      local len = line:len ()
      local pos, posEnd = line:find (spStartTokenOfUserCommandDefinition, 1, true)
      
      if not pos then
        haltWithError (lineNumber, fileName, line,
                        "Falscher Aufruf von \\meinBefehl",
                        "Der Befehl \\meinBefehl muss zu Beginn einer Zeile aufgerufen werden"..
                      "\\MessageBreak und zwischen \\\\space und meinBefehl darf keine Lücke sein."
                      )
        return false, true
      end
      
      pos = posEnd + 1
      if pos > len then
        haltWithError (lineNumber, fileName, line, 
                       generalSyntaxErrorForUserCommandsShort, 
                       generalSyntaxErrorForUserCommandsLong)
        return false, true
      end
      
      spCurrentlyParsedUserCommandParamCount = 0
      if line:sub (pos, pos) == "[" then
        pos = pos+1
        
        local nValue = nil
        
        local posZu = findFrom (line, "]", pos)
        if posZu > pos then
          local sValue = line:sub (pos, posZu-1):gsub(" ", "")
          nValue = tonumber (sValue)
        end
        
        if nValue == nil then
          haltWithError (lineNumber, fileName, line,
                          charBS.."meinBefehl erlaubt in []-Klammer nur ZIFFERN",
                          "Wenn du nach dem Befehl \\meinBefehl in eckigen Klammern die von dir erwartete"..
                        "\\MessageBreak Parameterzahl angibst, so mus dies eine ganze Zahl sein."..
                        "\\MessageBreak (Eine Null wäre auch erlaubt, doch statt [0] zu schreiben, wär's einfacher,"..
                        "\\MessageBreak\\space die [0] wegzulassen und gleich mit der öffnenden, geschweiften Klammer "..lBrace..
                        "\\MessageBreak\\space zu beginnen.)"
                        )
          return false, true
        end
        
        spCurrentlyParsedUserCommandParamCount = nValue
        pos = posZu+1
      end
      
      if (pos > len-2) or not (line:sub (pos, pos) == "{") then
        haltWithError (lineNumber, fileName, line,
                       generalSyntaxErrorForUserCommandsShort, 
                       generalSyntaxErrorForUserCommandsLong)
        return false, true
      end
      pos = pos+1
      
      local posZu = findFrom (line, "}", pos)
      if posZu < pos+1 then
        haltWithError (lineNumber, fileName, line,
                       generalSyntaxErrorForUserCommandsShort,
                       generalSyntaxErrorForUserCommandsLong)
        return false, true
      end
      
      spCurrentlyParsedUserCommandName = line:sub (pos, posZu-1):gsub(" ", "")
      
      --print ("|"..spCurrentlyParsedUserCommandName.."|")
      
      if spCurrentlyParsedUserCommandName:len () < 1 then
        haltWithError (lineNumber, fileName, line,
                       generalSyntaxErrorForUserCommandsShort,
                       generalSyntaxErrorForUserCommandsLong)
        return false, true
      end
      
      if spCurrentlyParsedUserCommandName:match ("%A") then
        haltWithError (lineNumber, fileName, line,
                        "Name von \\meinBefehl enthält verbotene Zeichen",
                        charBS.."meinBefehl erlaubt im Namen nur Zeichen aus dem lateinischen Alphabet,"..
                      "\\MessageBreak also nur Buchstaben, keine Umlaute, keine Ziffern und keine Sonderzeichen."
                      )
        return false, true
      end
      
      spCurrentlyInDefinitionOfUserCommand = true

      return true, true

    elseif startsWith (lineStripped, spEndTokenOfUserCommandDefinition) then
      
      storeUserCommand (spCurrentlyParsedUserCommandName, spCurrentlyParsedUserCommandParamCount, spRawLinesOfUserCommand)
      spCurrentlyInDefinitionOfUserCommand = false

      return true, true

    elseif spCurrentlyInDefinitionOfUserCommand then
      spRawLinesOfUserCommand [#spRawLinesOfUserCommand+1] = line

      return true, true
    end

    return true, false
  end


  function luaImportUserCommandFile (fileName)
    local rawFileExtra  = io.open (fileName, "r")
    
    local lineNumber = 0
    
    
    spCurrentlyInDefinitionOfUserCommand   = false
    spRawLinesOfUserCommand                = {}
    spCurrentlyParsedUserCommandName       = ""
    spCurrentlyParsedUserCommandParamCount = 0
        
    
    for line in rawFileExtra:lines () do
      lineNumber = lineNumber + 1
      --tex.print(line)
      --tex.print("")
      local lineStripped = line:gsub(" ", "")

      local bOK, bDummy = parseLineAndStoreUserCommandDefinitionIfAny (lineNumber, fileName, line, lineStripped)

      if not bOK then
        goto endOfFunction -- We need to call this goto because TeX won't process the \PackageError() from haltWithError before this method is done.
      end

    end

    ::endOfFunction::
    io.close (rawFileExtra)

    -- Check if last command definition was ended propperly:
    if spCurrentlyInDefinitionOfUserCommand then
      tex.print("\\PackageError{SPBuchsatz}"..
        "{Datei ist fehlerhaft: "..fileName..
        "\\MessageBreak\\space"..
        "\\MessageBreak Bitte beende den Textsatz mit X und Eingabetaste"..
        "}{"..
        "Bitte die X-Taste, dann die Eingabetaste."..
        "}"
      )
    end

    -- important: We sort the table according to command name lengths, longest first.
    --            So we can detect which \Befehl… is called by taking the first that's matching.
    --
    table.sort (spInternUserCommands, function (a, b) return a.name:len () > b.name:len () end)
      
  end
\end{luacode*}



% - Internal code for pre-processing document text lines - - - - - - - - - -

\begin{luacode*}
  
  function lineStartsProlog (lineStripped)
    return (
         startsWith (lineStripped, "\\Prolog[")
      or startsWith (lineStripped, "\\Prolog{")
      or startsWith (lineStripped, "\\PrologOhneHintergrundBild[")
      or startsWith (lineStripped, "\\PrologOhneHintergrundBild{")
      or startsWith (lineStripped, "\\PrologEinfach[")
      or startsWith (lineStripped, "\\PrologEinfach{")
      );
  end

  function lineStartsEpilog (lineStripped)
    return (
         startsWith (lineStripped, "\\Epilog[")
      or startsWith (lineStripped, "\\Epilog{")
      or startsWith (lineStripped, "\\EpilogOhneHintergrundBild[")
      or startsWith (lineStripped, "\\EpilogOhneHintergrundBild{")
      or startsWith (lineStripped, "\\EpilogEinfach[")
      or startsWith (lineStripped, "\\EpilogEinfach{")
      );
  end

  function lineStartsNormalChapter (lineStripped)
    return (
         startsWith (lineStripped, "\\Kapitel[")
      or startsWith (lineStripped, "\\Kapitel{")
      );
  end

  function lineStartsInlineChapter (lineStripped)
    return startsWith (lineStripped, "\\KapitelFortlaufend") -- also works for \KapitelFortlaufendMitBild
  end

  function lineStartsSpecialChapter (lineStripped)
    return (
         startsWith (lineStripped, "\\KapitelSpezial[")
      or startsWith (lineStripped, "\\KapitelSpezial{")
      );
  end

  function lineStartsStory (lineStripped)
    return (
         startsWith (lineStripped, "\\Geschichte[")
      or startsWith (lineStripped, "\\Geschichte{")
      or startsWith (lineStripped, "\\GeschichteMitBild[")
      or startsWith (lineStripped, "\\GeschichteMitBild{")
      );
  end

  function lineStartsStoryChapter (lineStripped)
    return (
         startsWith (lineStripped, "\\GeschichtenKapitel[")
      or startsWith (lineStripped, "\\Geschichtenkapitel[")
      or startsWith (lineStripped, "\\GeschichtenKapitel{")
      or startsWith (lineStripped, "\\Geschichtenkapitel{")
      );
  end
  

  function startsWithNormalChapterOrStory (lineStripped)
    return 
         lineStartsNormalChapter (lineStripped)  -- but no \KapitelSpezial here!
      or lineStartsStory (lineStripped)
  end

  
  function startsWithNormalChapterInlineChapterOrStory (lineStripped)
    return 
         startsWithNormalChapterOrStory (lineStripped)
      or lineStartsInlineChapter (lineStripped)
  end

  
  function startsWithPrologOrEpilog (lineStripped)
    return lineStartsProlog (lineStripped) or lineStartsEpilog (lineStripped)
  end

  
  function startsWithAnyChapterPrologEpilogOrStory (lineStripped)
    return 
         startsWithNormalChapterInlineChapterOrStory (lineStripped) 
      or startsWithPrologOrEpilog (lineStripped)
  end

  
  function foundEndOfScene (lineStripped)
    return startsWith (lineStripped, "\\Szenentrenner")
        or startsWith (lineStripped, "\\GeschichtenKapitel")
        or startsWith (lineStripped, "\\Geschichtenkapitel")
        or startsWith (lineStripped, "\\ZwischenTitel")
        or startsWith (lineStripped, "\\Zwischentitel")
        or startsWith (lineStripped, "\\Beitext")
        or startsWith (lineStripped, "\\vertikalDehnenUndSzene")
        or startsWith (lineStripped, "\\neueSeiteUndSzene")
        or startsWith (lineStripped, "\\VakatseiteEinsetzenUndSzene")
        or startsWith (lineStripped, "\\neueBildseiteUndSzene")
        or startsWith (lineStripped, "\\BildAbfallendObenUndSzene")
        or startsWith (lineStripped, "\\BildHoeheFixUndSzene")
        or startsWith (lineStripped, "\\BildHoeheDynUndSzene")
        or startsWith (lineStripped, "\\BildBreiteFixUndSzene")
        or startsWith (lineStripped, "\\BildBreiteDynUndSzene")
        or startsWithPattern (lineStripped, "\\Ende%u%a*UndSzene") 
  end
  
  
  -- This flag and function used when there are no chapters, no inline chapters,
  -- and no stories in Haupttext.tex
  -- So we can still enable Lombards / Initials according to Zentraldokument:
  -- 
  flagWeAreAtStartOfMaintextAndWeHaveNoChaptersNorStories = false
  
  function weAreAtStartOfMaintextAndWeHaveNoChaptersNorStories ()
    if flagWeAreAtStartOfMaintextAndWeHaveNoChaptersNorStories then
      flagWeAreAtStartOfMaintextAndWeHaveNoChaptersNorStories = false
      return true
    end
    return false
  end

  spInternLombardsActive = true

\end{luacode*}


\pruefeFeld{\spSzeneBeginntMitLombarde}
\ifFeldIstGesetzt
  \pruefeFeld{\spKapitelBeginntMitLombarde}
  \ifFeldIstGesetzt
    \begin{luacode*}
      function nextParNeedsLombarde (lineStripped)
        return spInternLombardsActive and 
          (    weAreAtStartOfMaintextAndWeHaveNoChaptersNorStories ()
            or foundEndOfScene (lineStripped) 
            or startsWithAnyChapterPrologEpilogOrStory (lineStripped) )
      end
    \end{luacode*}
  \else
    \begin{luacode*}
      function nextParNeedsLombarde (lineStripped)
        return spInternLombardsActive and 
          (    weAreAtStartOfMaintextAndWeHaveNoChaptersNorStories ()
            or foundEndOfScene (lineStripped) 
            or lineStartsStoryChapter (lineStripped)
            or startsWithPrologOrEpilog (lineStripped) )
      end
    \end{luacode*}
  \fi
\else
  \pruefeFeld{\spKapitelBeginntMitLombarde}
  \ifFeldIstGesetzt
    \begin{luacode*}
      function nextParNeedsLombarde (lineStripped)
        return spInternLombardsActive and 
          (    weAreAtStartOfMaintextAndWeHaveNoChaptersNorStories ()
            or startsWithNormalChapterInlineChapterOrStory (lineStripped) 
            or startsWithPrologOrEpilog (lineStripped) )
      end
    \end{luacode*}
  \else
    \begin{luacode*}
      function nextParNeedsLombarde (lineStripped)
        return false
      end
    \end{luacode*}
  \fi
\fi

\pruefeFeld{\spKapitelBeginntMitInitial}
\ifFeldIstGesetzt
  \begin{luacode*}
    function nextParNeedsInitial (lineStripped)
      return 
          (    weAreAtStartOfMaintextAndWeHaveNoChaptersNorStories ()
            or startsWithNormalChapterInlineChapterOrStory (lineStripped) )
    end
  \end{luacode*}
\else
  \begin{luacode*}
    function nextParNeedsInitial (lineStripped)
      return false
    end
  \end{luacode*}
\fi


\def\PrologVorverarbeitung{%
  \addtocounter{chapter}{-1}%
}

  
\def\spInternTxtFehltHierInitial{\textcolor{red}{\ Fehlt\\\ hier\\\textbackslash Ini-\\\ tial\{\}?}}
\newcommand{\spNoteFehltHierInitial}{%
\ifthenelse{\isodd{\value{page}}}{%
\reversemarginpar%
\marginpar{\raggedright\footnotesize{\spInternTxtFehltHierInitial}}%
\normalmarginpar%
}{%
\normalmarginpar%
\marginpar{\raggedright\footnotesize{\spInternTxtFehltHierInitial}}%
}}

\def\spInternTxtFehltHierLombarde{\textcolor{red}{\ Fehlt\\\ hier\\\textbackslash Lom-\\\ barde\{\}?}}
\newcommand{\spNoteFehltHierLombarde}{%
\ifthenelse{\isodd{\value{page}}}{%
\reversemarginpar%
\marginpar{\raggedright\footnotesize{\spInternTxtFehltHierLombarde}}%
\normalmarginpar%
}{%
\normalmarginpar%
\marginpar{\raggedright\footnotesize{\spInternTxtFehltHierLombarde}}%
}}

\def\spInternTxtAbsatzendeNachAbsNebenInitial{\rotatebox{90}{\schriftDejaVuSans{\klein\textcolor{red}{Fehler: Nach} \textcolor{DunkelRot}{\textbackslash AbsNebenInitial oder \textbackslash AbsNebenBild} \textcolor{red}{darf kein Absatzende sein.}}}}
\newcommand{\spNoteAbsatzendeNachAbsNebenInitial}{%
\ifthenelse{\isodd{\value{page}}}{%
\normalmarginpar%
\marginpar{\spInternTxtAbsatzendeNachAbsNebenInitial}%
\reversemarginpar%
}{%
\normalmarginpar%
\marginpar{\spInternTxtAbsatzendeNachAbsNebenInitial}%
}}



%  Automatische Vorverarbeitung der Eingabetexte:
%
%  Erzeugt die Hilfsdatei mainDocument.tex im aktuellem Verzeichnis, z. B. in ../build/
%  
%  Einfügen einer Leerzeile nach jedem Absatz,
%  verhindern von Erstzeileneinzug nach Kapitel- oder Szenewechsel,
%  ggfs. Einbauen von Initialen und/oder Lombarden,
%  ggfs. Escapen der reservierten Zeichen, 
%  ggfs. Anpassen der Nachverfolgung (Rot-/Grün-Markierungen)
%  um innerhalb der \alt-, \neu-, \altneu-Befehle beliebige
%  (auch zerbrechliche) Inhalte zu ermöglichen,
%   … und weitere Anpassungen.
%
%
%
\pruefeFeld{\spHalbenLeerschrittZwischenEllipseUndAusrufeOderFragezeichen}
\ifFeldIstGesetzt
  \begin{luacode*}
    insertHalfSpaceBetweetEllipseAndSpecialMarks = true
  \end{luacode*}
\else
  \begin{luacode*}
    insertHalfSpaceBetweetEllipseAndSpecialMarks = false
  \end{luacode*}
\fi
%
\begin{luacode*}
  function forbidCriticalOrphans (fileOut)
    fileOut:write("\n\\verbieteSchusterjungenNachLuecke\n\n")
  end

  function allowOrphans (fileOut, addExtraLF)
    if addExtraLF then
      fileOut:write("\n\\erlaubeSchusterjungenFallsAutomatikAktiv\n\n")
    else
      fileOut:write("\\erlaubeSchusterjungenFallsAutomatikAktiv\n")
    end
  end
  

  function ifFileExists (filename)
    print ("Looking for file: "..filename)
    local f=io.open (filename,"r")
    if f~=nil then
      io.close (f)
      tex.print (1)
      print ("File found: "..filename)
    else
      tex.print (0)
      print ("File not found: "..filename)
    end
  end

  
  function skipToken (line, token, iStart)
    local tokenLen = token:len ()
    if line:sub(iStart, iStart+tokenLen-1) == token then
      return iStart+tokenLen
    end
    return iStart
  end
  

  function skipInlineToken (line, iStart)
    local tokenStart = "\\inline"
    local tokenLen = tokenStart:len ()
    if line:sub(iStart, iStart+tokenLen-1) == tokenStart then
      local c = line:sub(iStart+tokenLen, iStart+tokenLen+2)
      if (c == "Etw") or (c == "Eng") or (c == "Wei") or (c == "Hal") then
        local posOpen = findFrom (line, "{", iStart+tokenLen+3)
        if posOpen > 0 then
          return posOpen+1
        end
      end
    end
    return iStart
  end
  

  function decorateCommandsInParameters (line, codeToReopen)
    local retVal = ""

    local lineLength = line:len()
    local lastIndex  = 0
    
    local idx=1
    repeat
      -- look for any \-command followed by a {…} part
      
      local posCommand = findFrom (line, "\\", idx)
      
      local commandLength  = 0
      local parseSecondToo = false
      
      if posCommand > 0 then
        local posOpen  = findFrom (line, "{", posCommand+1)
        local posClose = findFrom (line, "}", posCommand+1)
        if (posOpen > 0) and (posOpen < posClose) then
          if posCommand > 1 then
            retVal = retVal..line:sub (idx, posCommand-1)
          end
          
          --close the sensitive command and re-open if behind the {
          retVal = retVal.."}"
          retVal = retVal..line:sub (posCommand, posOpen)
          retVal = retVal..codeToReopen
          retVal = retVal..line:sub (posOpen+1, posClose-1)
          retVal = retVal.."}}"..codeToReopen
          
          idx = posClose
          
          lastIndex = idx
        end
      end
      idx = idx+1
    until (idx > lineLength)

    retVal = retVal..line:sub (lastIndex+1)
    return retVal
  end

  
  function codiereEinenSchalter (line, spToken, newTokenA, newTokenB, codeToReopenA, codeToReopenB)
    local retVal = ""
    
    local posFirst = findFirst (spToken)
    if posFirst == 0 then
      return line
    end

    local skipFirst = (posFirst > 1)
    local parts = splitString (line, spToken)
    for part in parts do
      if skipFirst then
        retVal = part
        skipFirst = false
      else 
        if part:len () > 0 then  -- ignore empty strings to work around a bug in splitString ()

          local posClose = findClosingBrace (part)
          if posClose == 0 then
            retVal = retVal..spToken..part
          else
            if hasToken (part, "\\") then
              local innerPart    = part:sub (1, posClose)
              local trailingPart = part:sub (posClose+1)
              
              retVal = retVal..newTokenA..decorateCommandsInParameters(innerPart, codeToReopenA)
              
              if startsWith (trailingPart, "{") and not (codeToReopenB == "") then
                retVal = retVal..newTokenB
                trailingPart = trailingPart:sub (2)
                local posClose = findClosingBrace (trailingPart)
                if posClose > 0 then
                  local innerPart         = trailingPart:sub (1, posClose)
                  local finalTrailingPart = trailingPart:sub (posClose+1)
              
                  trailingPart = decorateCommandsInParameters(innerPart, codeToReopenB)..finalTrailingPart
                end
              end
              retVal = retVal..trailingPart
            else
              retVal = retVal..spToken..part
            end
          end
          
        end
      end
    end
    
    return retVal
  end

  
  function entferneKommando (lineRaw, spTokenWithoutBrace, removeFirstParameter)
    local retVal = ""
    
    local posFirst = findFirst (lineRaw, spTokenWithoutBrace)
    if posFirst == 0 then
      return lineRaw
    end

    local spToken        = spTokenWithoutBrace .. "{"
    local spTokenWithGap = spTokenWithoutBrace .. "%s*{"

    local line = lineRaw:gsub(spTokenWithGap, spToken)
    
    posFirst = findFirst (line, spToken)
    if posFirst == 0 then
      return lineRaw
    end

    local skipFirst = (posFirst > 1)
    local parts = splitString (line, spToken)
    for part in parts do
      if skipFirst then
        retVal = part
        skipFirst = false
      else 
        if (part:len () > 0) then  -- ignore empty strings to work around a bug in splitString ()

          local posClose = findClosingBrace (part)
          
          if posClose == 0 then
            -- do nothing: we ignore this part and do not return it
          else
            if removeFirstParameter then
              retVal = retVal..part:sub (posClose+1)
            else
              retVal = retVal..part:sub (1, posClose-1)..part:sub (posClose+1)
            end
          end
        end
      end
    end
   
    return retVal
  end

  
  function codiereAltNeuSchalter (line)
    if spInternNachverfolgungsModusAktiv 
    then
      local retVal = codiereEinenSchalter (
        line, "\\altneu{", "\\altneu[x]{}{}\\alt[x]{", "\\neu[x]{", 
        "\\alt[x]{", "\\neu[x]{")
        --"\\RotDurchgestrichen{", "\\GruenDickUnterstrichen{")
      
      retVal = codiereEinenSchalter (
        retVal, "\\alt{", "\\alt{", "", 
        "\\alt[x]{", "")
      
      retVal = codiereEinenSchalter (
        retVal, "\\neu{", "\\neu{", "", 
        "\\neu[x]{", "")

      return retVal
    else
      local retVal 
             = entferneKommando (line,   "\\altneu", true)
      retVal = entferneKommando (retVal, "\\alt",    true)
      retVal = entferneKommando (retVal, "\\neu",    false)
      
      return retVal
    end
  end
  

  currentChapter = 0
  currentPart    = 0
  skipErstzeilenAutomatikGewuenscht  = false
  restauriereClubPenalty             = false
  mustDecFlagParagraphIsRightAligned = false

  setzeEinInitial   = false
  setzeEineLombarde = false


  totalNumberOfPrologs        = 0
  totalNumberOfNormalChapters = 0
  totalNumberOfInlineChapters = 0
  totalNumberOfStories        = 0
  totalNumberOfPrologsNormalChaptersInlineChaptersStories = 0  -- does NOT count \GeschichtenKapitel
  
  spTmpIgnoreIsActive = false

  
  function processPrologChapterStoryCommands (lineInRaw)
    local lineStripped = lineInRaw:gsub(" ", "")

    if startsWith (lineStripped, "%") then
      return
    end
    
    if lineStartsProlog (lineStripped)
    then
      totalNumberOfPrologsNormalChaptersInlineChaptersStories = totalNumberOfPrologsNormalChaptersInlineChaptersStories + 1
      tex.print ("\\PrologVorverarbeitung")
      totalNumberOfPrologs = totalNumberOfPrologs + 1
      currentChapter = currentChapter-1
      --io.write ("currentChapter: "..currentChapter.."  ")
    elseif lineStartsNormalChapter (lineStripped)
    then
      totalNumberOfPrologsNormalChaptersInlineChaptersStories = totalNumberOfPrologsNormalChaptersInlineChaptersStories + 1
      totalNumberOfNormalChapters = totalNumberOfNormalChapters + 1
    elseif lineStartsInlineChapter (lineStripped)
    then
      totalNumberOfPrologsNormalChaptersInlineChaptersStories = totalNumberOfPrologsNormalChaptersInlineChaptersStories + 1
      totalNumberOfInlineChapters = totalNumberOfInlineChapters + 1
    elseif lineStartsStory (lineStripped)
    then
      totalNumberOfPrologsNormalChaptersInlineChaptersStories = totalNumberOfPrologsNormalChaptersInlineChaptersStories + 1
      totalNumberOfStories = totalNumberOfStories + 1
    end
  end
  
  function preprocessLine (lineInRaw)
    local lineIn = lineInRaw:gsub(string.char(0x09), "")                      --remove TAB
                            :gsub(string.char(0x0c), "")                      --remove FORM FEED
                            :gsub(string.char(0xC2)..string.char(0xA0), "~")  --replace NO-BREAK SPACE
    -- and finally replace
    -- any explicit extra SPACE
    -- by its robust version:
                            :gsub("%\\\\ ", "!!Zeilenumbruch!! ")
                            :gsub("%\\ ", "\\Leerschritt ")
                            :gsub("!!Zeilenumbruch!! ", "\\\\ ")

    if insertHalfSpaceBetweetEllipseAndSpecialMarks then
      lineIn = lineIn:gsub("%…%!", "…\\,!")
      lineIn = lineIn:gsub("%…%?", "…\\,?")
    end

    local lineStripped = lineIn:gsub(" ", "")

    local isComment = startsWith (lineStripped, "%")
    
    return lineIn, lineStripped, isComment
  end

  
  local tokenStartOfMeinBefehl = "\\def\\spTmpIgnoriereBefehl{"
  local tokenEndOfMeinBefehl   = "}%spTmpIgnoriereBefehl"

  weAreInAbsatztabelle            = false
  waitingForClosingOfCommandLevel = 0
  
  
  function startsWithCommandThatMustBeProcessedAlways (lineStripped)
    return startsWith (lineStripped, "\\setzeKapitel") or
           startsWith (lineStripped, "\\ladeVorlageD") or
           startsWith (lineStripped, "\\meinBefehl") or
           startsWith (lineStripped, "\\def\\") or 
           startsWith (lineStripped, "\\if") or 
           startsWith (lineStripped, "\\else") or 
           startsWith (lineStripped, "\\fi") or 
           startsWith (lineStripped, "\\setzeInhalte")
  end
  
  
  function appendFixedLineToMainDocument (lineIn, lineStripped, fileOut, oldWeAreInTabular, addExtraLF, wantedChapterIdx, wantedChapterIdxLast, wantedPartIdx, globalNoIndent, fileOutForRTF)

    local lineOutForRTF = lineIn

    if mustDecFlagParagraphIsRightAligned then
      fileOut:write("\\decFlagParagraphIsRightAligned\n")
      mustDecFlagParagraphIsRightAligned = false
    end
    
    if lineIn:match('\\deaktiviereLombarden') then
      spInternLombardsActive = false
      setzeEineLombarde = false
      lineIn       =       lineIn:gsub("%\\deaktiviereLombarden", "")
      lineStripped = lineStripped:gsub("%\\deaktiviereLombarden", "")
    else
      if lineIn:match('\\aktiviereLombarden') then
        spInternLombardsActive = true
        lineIn       =       lineIn:gsub("%\\aktiviereLombarden", "")
        lineStripped = lineStripped:gsub("%\\aktiviereLombarden", "")
      end
    end
    
    local allChapters = (wantedChapterIdx == -99)
    local allParts    = (wantedPartIdx    == -99)

    local isStartOfPart = startsWith (lineStripped, "\\Buchteil{")

    local isStartOfPrologOrEpilog = lineStartsProlog (lineStripped) or lineStartsEpilog (lineStripped)
    local isStartOfChapterOrStory = startsWithNormalChapterOrStory (lineStripped)
    local isStartOfInlineChapter  = lineStartsInlineChapter (lineStripped)
    local isStartOfStoryChapter   = lineStartsStoryChapter (lineStripped)
    
    local isStartOfInlineChapterOrStoryChapter = isStartOfInlineChapter or isStartOfStoryChapter


    if    isStartOfPart
       or isStartOfChapterOrStory
       or isStartOfInlineChapterOrStoryChapter
       or isStartOfPrologOrEpilog
    then
      spInternLombardsActive = true
    end

    -- Abschnitte, nach denen (1.) nicht eingerückt wird und (2.) kein Schusterjunge sein darf
    local neuerTeilAbschnitt = 
         isStartOfChapterOrStory
      or isStartOfInlineChapterOrStoryChapter
      or isStartOfPrologOrEpilog
      or startsWith (lineStripped, "\\Szenentrenner")
      or startsWith (lineStripped, "\\BeginnEinrueckung");
    
    -- erkenne weitere Abschnitte, nach denen kein Schusterjunge sein darf
    local mustPreventFollowingOrphan = neuerTeilAbschnitt
      or startsWith (lineStripped, "\\EndeEinrueckung")
      or startsWith (lineStripped, "\\BeginnLinks")
      or startsWith (lineStripped, "\\EndeLinks")
      or startsWith (lineStripped, "\\BeginnZentriert")
      or startsWith (lineStripped, "\\EndeZentriert")
      or startsWith (lineStripped, "\\BeginnRechts")
      or startsWith (lineStripped, "\\EndeRechts")
      or startsWith (lineStripped, "\\BeginnKursiv")
      or startsWith (lineStripped, "\\EndeKursiv")
      or startsWith (lineStripped, "\\BeginnFett")
      or startsWith (lineStripped, "\\EndeFett")
      or startsWith (lineStripped, "\\BeginnKapitaelchen")
      or startsWith (lineStripped, "\\EndeKapitaelchen")
      or startsWith (lineStripped, "\\BildHoehe")
      or startsWith (lineStripped, "\\BildBreite")
      or startsWith (lineStripped, "\\BildAbfallend")
      or startsWith (lineStripped, "\\Leerzeile")
      or startsWith (lineStripped, "\\EndeKleintabelle")
      or startsWith (lineStripped, "\\EndeTabelle")
      ;
    
    
    -- Special handling of commands that need to be processed, even when they
    -- are not in the chapter (or in the part, resp.) that we want to process:
    -- 
    local thisLineStartsWithCommandThatMustBeProcessedAlways = false
    
    
    if not allChapters then
      if isStartOfChapterOrStory or isStartOfPrologOrEpilog
      then
        currentChapter = currentChapter+1
      end

      local wantThisChapter = 
        ((currentChapter == wantedChapterIdx) or
        ((not (wantedChapterIdxLast == nil)) and (currentChapter > wantedChapterIdx) and (currentChapter <= wantedChapterIdxLast)))
      
      if isStartOfChapterOrStory or isStartOfPrologOrEpilog
      then
        if wantThisChapter
        then
          if skipErstzeilenAutomatikGewuenscht
          then
            skipErstzeilenAutomatikGewuenscht = false
            fileOut:write("\\DiesmalKeineErstzeilenAutomatik\n")
          end
        else
          fileOut:write("\n%%% Ersatzbefehl anstelle des übersprungenen Kapitels:\n")
          fileOut:write("\\addtocounter{chapter}{1}\n\n")
        end
      end
      
      if not wantThisChapter then
        if     (not isStartOfPart)
           and (waitingForClosingOfCommandLevel == 0)
        then
          if startsWithCommandThatMustBeProcessedAlways (lineStripped) then
            thisLineStartsWithCommandThatMustBeProcessedAlways = true
          else
            if  neuerTeilAbschnitt then
              skipErstzeilenAutomatikGewuenscht = false
            else
              if startsWith (lineStripped, "\\DiesmalKeineErstzeilenAutomatik") then
                skipErstzeilenAutomatikGewuenscht = true
              end
            end
            return oldWeAreInTabular
          end
        end
      end
    end
    
    if not allParts then
      if isStartOfPart then
        currentPart = currentPart+1
        if (currentPart == wantedPartIdx) and 
          skipErstzeilenAutomatikGewuenscht
        then
          skipErstzeilenAutomatikGewuenscht = false
          fileOut:write("\\DiesmalKeineErstzeilenAutomatik\n")
        end
      end
      if not (currentPart == wantedPartIdx) then
        if     (not isStartOfPart)
           and (waitingForClosingOfCommandLevel == 0)
        then
          if startsWithCommandThatMustBeProcessedAlways (lineStripped) then
            thisLineStartsWithCommandThatMustBeProcessedAlways = true
          else
            if  neuerTeilAbschnitt then
              skipErstzeilenAutomatikGewuenscht = false
            else
              if startsWith (lineStripped, "\\DiesmalKeineErstzeilenAutomatik") then
                  skipErstzeilenAutomatikGewuenscht = true
              end
            end
            if   isStartOfChapterOrStory
              or isStartOfPrologOrEpilog 
              or isStartOfInlineChapter
            then
              fileOut:write("\n%%% Ersatzbefehl anstelle des übersprungenen Kapitels:\n")
              fileOut:write("\\addtocounter{chapter}{1}\n\n")
            end
            return oldWeAreInTabular
          end
        end
      end
    end



    if thisLineStartsWithCommandThatMustBeProcessedAlways then
      fileOut:write("\n%%% Dieses Kommando wird berücksichtigt, obwohl das aktuelle Kapitel übersprungen wird:\n")
      local posOpen = findFrom (lineStripped, "{", 3)
      if posOpen > 0 then
        local partAfterOpen = lineStripped:sub (posOpen+1)
        local posClose, remainingDepth = findClosingBraceAndRemainingDepth (partAfterOpen)
        if posClose == 0 then
          waitingForClosingOfCommandLevel = waitingForClosingOfCommandLevel + remainingDepth
--fileOut:write("\n%%% Tiefe der Bereichschachtelung:",waitingForClosingOfCommandLevel,"\n")
        end
      end
    elseif waitingForClosingOfCommandLevel > 0 then
      local posClose, remainingDepth = findClosingBraceAndRemainingDepth (lineStripped)
      if posClose > 0 then
        waitingForClosingOfCommandLevel = waitingForClosingOfCommandLevel - 1
        if waitingForClosingOfCommandLevel < 0 then
          waitingForClosingOfCommandLevel = 0
        end
      else
        waitingForClosingOfCommandLevel = waitingForClosingOfCommandLevel + remainingDepth - 1
--fileOut:write("\n%%% Tiefe der Bereichschachtelung:",waitingForClosingOfCommandLevel,"\n")
      end
    end
    
    
    
    -- Buchteil und Kapitel stimmen (oder immer-verarbeiten-Befehl gefunden), diese Zeile wird verarbeitet:
    
    if lineStripped:match('\\horizontalDehnen') then
      fileOut:write("\\incFlagParagraphIsRightAligned\n")
      mustDecFlagParagraphIsRightAligned = true
    end

    if startsWithBeginParTableCommand (lineStripped) then
      weAreInAbsatztabelle = true
    elseif startsWithEndParTableCommand (lineStripped) then
      weAreInAbsatztabelle = false
    end

    local weAreInTabular = oldWeAreInTabular or
      lineStripped:match('\\BeginnKleintabelle') or
      lineStripped:match('\\begin{tabu') or
      lineStripped:match('\\begin{longtable') or
      weAreInAbsatztabelle

    local lineOut = lineIn:gsub("%$", "\\$")
    
    if startsWithAnyChapterPrologEpilogOrStory (lineStripped) then
      lineOut = lineOut:gsub("%\\kursiv{",  "\\kursivInAnyChapterPrologEpilogOrStory{")
      lineOut = lineOut:gsub("%\\kursiv {", "\\kursivInAnyChapterPrologEpilogOrStory{")
    end
    
    if not startsWith (lineStripped, "\\meinBefehl") then
      lineOut = lineOut:gsub("%#", "\\#")
    end
    
    local extraLine = ""

    if not weAreInTabular then
      lineOut = lineOut:gsub("%&", "\\&")
    end


    
    if globalNoIndent and (lineOut:len () > 0) then
        fileOut:write("\\noindent ")
    end
    
    if not weAreInTabular then
      local pos = findFirst (lineOut, '\\BeginnEinrueckung')
      if pos > 0 then
        local posComment = findFrom (lineOut, '%', pos+18)
        if posComment > 0 then
          lineOut = lineOut:sub(1, posComment-1)
        end
      --  local posBackslash = findFrom (lineOut, '\\', pos+18)
      --  if posBackslash > 0 then
      --    extraLine = lineOut:sub(posBackslash)
      --    lineOut   = lineOut:sub(1, posBackslash-1)
      --  end
      end
    end

    
    local startOfTmpIgnore = (lineStripped == tokenStartOfMeinBefehl)
    local endOfTmpIgnore   = (lineStripped == tokenEndOfMeinBefehl)
    
    if startOfTmpIgnore then
      spTmpIgnoreIsActive = true
    end

    if not spTmpIgnoreIsActive then
      if    (setzeEinInitial or setzeEineLombarde)
        and lineStripped 
        and (lineStripped:len () > 0)
        and not (
                     startsWith (lineStripped, "\\BeginnEng")
                  or startsWith (lineStripped, "\\BeginnHalbEng")
                  or startsWith (lineStripped, "\\BeginnHalbWeit")
                  or startsWith (lineStripped, "\\BeginnWeit")
                  or startsWith (lineStripped, "\\BeginnLaufweite")
                  or startsWith (lineStripped, "\\BeginnWortabstand")
                  or startsWith (lineStripped, "\\BeginnKursiv")
                  or startsWith (lineStripped, "\\KeineSeitenzahl")
                  or startsWith (lineStripped, "\\MitSeitenzahl")
                  or startsWith (lineStripped, "\\KeinRBild")
                  or startsWith (lineStripped, "\\KeineRBilder")
                  or startsWith (lineStripped, "\\MitRBild")
                  or startsWithPattern (lineStripped, "\\Bild%u%a*Fix") 
                  or startsWithPattern (lineStripped, "\\Bild%u%a*Dyn") 
                  or startsWith (lineStripped, "\\neueBildseite")
                  or startsWith (lineStripped, "\\Beitext")
                  or startsWith (lineStripped, "\\ZusatzzeileImIV")
                  or startsWith (lineStripped, "\\Leerzeile")
                  or startsWith (lineStripped, "\\ZweiLeerzeilen")
                  or startsWith (lineStripped, "\\DreiLeerzeilen")
                  or startsWith (lineStripped, "\\VierLeerzeilen")
                  or startsWith (lineStripped, "\\FaktorLeerzeile")
                  or startsWith (lineStripped, "\\GeheZeileNachOben")
                  or startsWith (lineStripped, "\\GeheZuRegisterzeile")
                  or startsWith (lineStripped, "\\RandnotizTW")
                  or startsWith (lineStripped, "\\RandnotizTH")
                  or startsWith (lineStripped, "\\RandnotizT")
                  or startsWith (lineStripped, "\\RandnotizCN")
                  or startsWith (lineStripped, "\\Notiz") 
                  or startsWith (lineStripped, "\\vertikalDehnen") 
                  or startsWith (lineStripped, "\\ladeVorlageDatei") 
                  or startsWith (lineStripped, "\\loescheGeschichteKopfzeile") 
                  or startsWith (lineStripped, "\\KopfzeileEinmalAuslassen") 
                  or startsWith (lineStripped, "\\setzeMarke")
                  or startsWith (lineStripped, "\\setzeInhalte")
                  or startsWith (lineStripped, "\\aktiviereLombarden")
                  or startsWith (lineStripped, "\\deaktiviereLombarden")
                  or startsWith (lineStripped, "\\ErsterAbsatzNachKapitelOderSzeneBeginntrue")
                  or startsWith (lineStripped, "\\Szenentrenner")
                  or startsWith (lineStripped, "\\pruefe")
                  or startsWith (lineStripped, "\\if")
                  or startsWith (lineStripped, "\\def\\")
                  or startsWith (lineStripped, "\\else")
                  or startsWith (lineStripped, "\\fi")
                  or startsWith (lineStripped, "}")
                  -- also ignore all of the \setze… commands:
                  or startsWithPattern (lineStripped, "\\setze%u%a*") 
                )
      then
        local setzeEinInitialWarGewuenscht = setzeEinInitial
      
        local i    = 1
        local iMax = lineOut:len ()
        local bFalseGlyphAtStart = false
        
        local hasUserInitial  = hasToken (lineOut, '\\Initial{')
        local hasUserLombarde = hasToken (lineOut, '\\Lombarde{')

        
        -- Folgt kein normaler Absatz, sondern zentriert, rechtsbündig,
        -- linksbündig oder Kapitälchen, so wird auf das Initial bzw.
        -- auf die Lombarde verzichtet:
        -- Es bzw. sie wird also nicht später nachgeholt, sondern entfällt.
        -- 
        -- Grund: Wollten Leute, dass das Initial bzw. die Lombarde später
        --        nachgeholt wird, so hätten sie \Beitext verwendet.
        -- 
        if     hasUserInitial
            or hasUserLombarde
            or startsWith (lineStripped, "\\BeginnZentriert") 
            or startsWith (lineStripped, "\\BeginnLinks")
            or startsWith (lineStripped, "\\BeginnRechts")
            --or startsWith (lineStripped, "\\BeginnKursiv")
            or startsWith (lineStripped, "\\kapitaelchen{") 
            or (     setzeEinInitial   
                and (    startsWith (lineStripped, "\\keinInitial")   
                      or startsWith (lineStripped, "\\keinEinzug\\keinInitial")
                    ) 
              )
            or (     setzeEineLombarde 
                and (    startsWith (lineStripped, "\\keineLombarde") 
                      or startsWith (lineStripped, "\\keinEinzug\\keineLombarde")
                    )
              )
        then
        
          if setzeEinInitial then
            setzeEinInitial = false
          else
            setzeEineLombarde = false
          end
          
          if (not hasUserInitial) and not hasUserLombarde then
            local pos = false
            local tokenLen = 0
            if setzeEinInitial then
              pos = findFirst (lineOut, '\\keinInitial')
              tokenLen = 12
            else
              pos = findFirst (lineOut, '\\keineLombarde')
              tokenLen = 14
            end
            if pos > 0 then
              -- den Befehl wegschneiden:
              local leadingText = ""
              if pos > 1 then
                leadingText = lineOut:sub(1, pos-1)
              end
              lineOut = leadingText..lineOut:sub(pos+tokenLen) 
            end
          end
          
        else
          while i <= iMax do

            i = skipToken (lineOut, " ", i)
            i = skipInlineToken (lineOut, i)
            i = skipToken (lineOut, " ", i)
            i = skipToken (lineOut, "\\keinEinzug", i)
            i = skipToken (lineOut, " ", i)
            i = skipToken (lineOut, "\\kursiv{",    i)
            i = skipToken (lineOut, " ", i)
            i = skipToken (lineOut, "\\fett{",      i)
            i = skipToken (lineOut, " ", i)
            
            -- nochmals nach kursiv schauen, um nicht nur
            -- \kursiv\fett zu erkennen, sondern auch \fett\kursiv
            i = skipToken (lineOut, "\\kursiv{",    i)
            i = skipToken (lineOut, " ", i)
            
            -- und vorsichtshalber nochmals nach \\inline… schauen:
            i = skipInlineToken (lineOut, i)
            i = skipToken (lineOut, " ", i)
            
            local newLineOut = ""
            if i > 1 then
              newLineOut = lineOut:sub(1, i-1)
            end
            
            if i <= iMax then
              local c = lineOut:sub(i, i)
              if c == "\\" then
                newLineOut = lineOut
                setzeEinInitial    = false
                setzeEineLombarde  = false
                bFalseGlyphAtStart = true
              else
                local testLine = lineOut:sub(i)
                
                local preInitialLine = ""
                
                for strGlyph in testLine:utfcharacters () do
                  if bFalseGlyphAtStart then
                    newLineOut = newLineOut..strGlyph
                  elseif setzeEinInitial or setzeEineLombarde then
                    if   (strGlyph == " ")                          
                      or (("‘" <= strGlyph) and (strGlyph <= "‟")) -- acht alte, einfache, doppelte und gespiegelte Anführungszeichen
                      or (strGlyph == "«") or (strGlyph == "»")       
                      or (strGlyph == "‹") or (strGlyph == "›") 
                      or (strGlyph == "¡") or (strGlyph == "¿") 
                    then
                      if setzeEinInitial then
                        preInitialLine = preInitialLine..strGlyph
                      else
                        -- skip this char, we leave it unchanged
                        newLineOut = newLineOut..strGlyph
                      end
                    else
                      if   (("A" <= strGlyph) and (strGlyph <= "Z"))
                        or (isUmlaut(strGlyph))
                      then
                        if setzeEinInitial then
                          if preInitialLine:len () > 0 then
                            newLineOut = newLineOut.."\\Initial["..preInitialLine.."]{"..strGlyph.."}"
                          else
                            newLineOut = newLineOut.."\\Initial{"..strGlyph.."}"
                          end
                        else
                          newLineOut = newLineOut.."\\Lombarde{"..strGlyph.."}"
                        end
                        setzeEinInitial   = false
                        setzeEineLombarde = false
                      else
                        newLineOut = newLineOut..strGlyph
                        setzeEinInitial    = false
                        setzeEineLombarde  = false
                        bFalseGlyphAtStart = true
                      end
                    end
                  else
                    newLineOut = newLineOut..strGlyph
                  end
                end   -- for
              end  -- if
              i = iMax+1
            end
            lineOut = newLineOut
          end
        end
        
        --print ("\n\nnew lineOut:", lineOut, "\n\n")
        
        -- Store
        
        if bFalseGlyphAtStart and spInternFeinkorrekturModusAktiv then
          if setzeEinInitialWarGewuenscht then
            fileOut:write("\n\\spNoteFehltHierInitial ")
          else
            fileOut:write("\n\\spNoteFehltHierLombarde ")
          end
        end
      else
        local pos = findFirst (lineOut, '\\keinInitial')
        if pos > 0 then
          local leadingText = ""
          if pos > 1 then
            leadingText = lineOut:sub(1, pos-1)
          end
          lineOut = leadingText..lineOut:sub(pos+12) -- den Befehl wegschneiden
        end
        local pos = findFirst(lineOut, '\\keineLombarde')
        if pos > 0 then
          local leadingText = ""
          if pos > 1 then
            leadingText = lineOut:sub(1, pos-1)
          end
          lineOut = leadingText..lineOut:sub(pos+14) -- den Befehl wegschneiden
        end
      end
      
      local posLombardeForRTF = findFirst (lineOut, '\\Lombarde{')
      if posLombardeForRTF == 0 then
        posLombardeForRTF = findFirst (lineOut, '\\Lombarde ')
      end
      if posLombardeForRTF == 0 then
        posLombardeForRTF = findFirst (lineOut, '\\Initial{')
      end
      if posLombardeForRTF == 0 then
        posLombardeForRTF = findFirst (lineOut, '\\Initial ')
      end
      if posLombardeForRTF > 0 then
        lineOutForRTF = lineOut
      end
      fileOutForRTF:write(lineOutForRTF.."\n")
      
      --local lineUpper = string.upper(lineOut)
      local noFileNamesHere = not (    lineStripped:match('\\Bild',                                1, true)
                                    or lineStripped:match('\\neueBildseite',                       1, true)
                                    or lineStripped:match('\\KapitelFortlaufendMitBild',           1, true)
                                    or lineStripped:match('\\Geschichte{',                         1, true)
                                    or lineStripped:match('\\GeschichteMitBild{',                  1, true)
                                    or lineStripped:match('\\ladeVorlageDatei{',                   1, true)
                                    or lineStripped:match('\\setzeKapitelTitelbild{',              1, true)
                                    or lineStripped:match('\\setzeKapitelTitelHintergrundGrafik{', 1, true)
                                    or lineStripped:match('\\setzeBuchteilTitelbildOben',          1, true)
                                    or lineStripped:match('\\setzeBuchteilTitelbildUnten',         1, true)
                                    or lineStripped:match('\\setzeSzenentrennerBild',              1, true)
                                    or lineStripped:match('\\setzeRandbild',                       1, true)
                                    or lineStripped:match('\\setzeRahmenbild',                     1, true)
                                  )
                              
      if noFileNamesHere then
        lineOut = lineOut:gsub("%_", "\\_")
        if spInternUsingSpecialFAE then
          lineOut = makeFineFAEs (lineOut)
        end
      end

      lineOut = codiereAltNeuSchalter (lineOut)

      fileOut:write(lineOut.."\n")

      if endsWith (lineStripped, "\\AbsNebenInitial") or endsWith (lineStripped, "\\AbsNebenBild") then
        fileOut:write("\\spNoteAbsatzendeNachAbsNebenInitial\n")
      end
      
      if lineOut:len () > 0 then
        lineNr = lineNr + 1
        storeDebugLineNumber (fileOut, lineNr)
      end
      
      if addExtraLF and not weAreInTabular then
        fileOut:write("\n")
      end
      
      if extraLine:len () > 0 then
        fileOut:write(extraLine.."\n")
        if addExtraLF and not weAreInTabular then
          fileOut:write("\n")
        end
        print ("")
        print (" Text hinter \\BeginnEinrueckung gefunden und in neue Zeile verschoben")
      end
   
    
      if mustPreventFollowingOrphan
      then
        forbidCriticalOrphans (fileOut)
        restauriereClubPenalty = true
      elseif restauriereClubPenalty 
        and (not weAreInTabular)
        and (lineOut:len () > 0) 
        and (    startsWith (lineStripped, "\\keinEinzug") 
              or startsWith (lineStripped, "\\kursiv") 
              or startsWith (lineStripped, "\\fett") 
              or startsWith (lineStripped, "\\inline") 
              or (not startsWith (lineStripped, "\\"))) 
      then
        allowOrphans (fileOut, addExtraLF)
        restauriereClubPenalty = false
      end
    else
      if mustPreventFollowingOrphan
      then
        forbidCriticalOrphans (fileOut)
        restauriereClubPenalty = true
      end
    end
    
    
    if     nextParNeedsInitial  (lineStripped) then
      setzeEinInitial = true
    elseif nextParNeedsLombarde (lineStripped) then
      setzeEineLombarde = true
    end
            
            
    if endOfTmpIgnore then
      spTmpIgnoreIsActive = false
    end

    
    return weAreInTabular and not
      ( lineStripped:match('\\EndeKleintabelle') or
        lineStripped:match('\\EndeTabelle') or
        lineStripped:match('\\end{tabu') or
        lineStripped:match('\\end{longtable') )
  end

  
  
  local function split (str)
    local pat = "[^-]*"    -- splitting at every - sign
    local tbl = {}
    str:gsub(pat, function(x) tbl[#tbl+1]=x end)
    return tbl
  end


  
  function file_exists (file)
    local f = io.open (file, "rb")
    if f then 
      f:close () 
    end
    return f ~= nil
  end



  -- for Haupttext.tex and Schlussseiten.tex

  
  function replaceParametersInCommandContent (line, paramCount, parameters)
    local result = line
    for idx = paramCount,1, -1 do
      local sIdx = tostring (idx)
      result = result:gsub ("#"..sIdx, parameters [idx])
    end
    return result
  end
  
  
  local resultingLinesOfPreRead = {}


  function buildParameterError (lineNumber, commandName, details)
    local sLineNumber = tostring (lineNumber)
    return {
      "\\PackageError{SPBuchsatz}"..
      "{Fehler in Zeile "..sLineNumber..":"..
        "\\MessageBreak\\space"..
        "\\MessageBreak Falscher Aufruf von Befehl"..commandName.."."..
        "\\MessageBreak Details: "..details..
        "\\MessageBreak\\space"..
        "\\MessageBreak\\space"..
        "\\MessageBreak Bitte beende Textsatz mit X und Eingabetaste"..
      "}{Bitte die X-Taste, dann die Eingabetaste.}"
    }
  end
  
  
  function getLinesWithExpandedCommands (lineNumber, fileName, line)
    local newLines = {}

    local lineStripped = line:gsub(" ", "")
    local isComment = startsWith (lineStripped, "%")
    
    local pos = 0
    if not isComment then
      pos = findPatternFirst (line, "\\Befehl%a")
    end
    
    if pos > 0 then

      local part = line:sub (pos + 7)

      local command = findBestMatchingUserCommand (lineNumber, fileName, part)

      if command then
        parameters = {}
        local leadingPart  = ""
        local trailingPart = ""
        
        local sExpectedParamCount = tostring (command.paramCount)
        local sFehltOderFehlen    = "fehlen"
        
        if sExpectedParamCount == "1" then
          sExpectedParamCount   = "Ein"
          sFehltOderFehlen      = "fehlt"
        end
              
        if pos > 1 then
          leadingPart = line:sub (1, pos-1)
        end
  
        if command.paramCount == 0 then
        
          if part:len () > command.name:len () then
            trailingPart = part:sub (command.name:len ()+1)
          end
          
        else
          local posB = 1
          
          for paramIdx = 1, command.paramCount do
            local posOpen  = findFrom (part, "{", posB)
            if posOpen == 0 then
              if paramIdx == 1 then
                return true, true, buildParameterError (lineNumber, command.name, sExpectedParamCount.." {}-Parameter "..sFehltOderFehlen..".")
              else
                local sIdx = tostring (paramIdx-1)
                local sFound = "keiner"
                if paramIdx > 1 then
                  sFound = "nur "..sIdx
                end
                return true, true, buildParameterError (lineNumber, command.name, sExpectedParamCount.." {}-Parameter erwartet, "..sFound.." gefunden.")
              end
            end
            
            local partAfterOpen = part:sub (posOpen+1)
            local posClose = findClosingBrace (partAfterOpen)
            if posClose == 0 then
              local sIdx = tostring (paramIdx)
              return true, true, buildParameterError (lineNumber, command.name, "Klammer fehlt nach "..sIdx..". {}-Parameter.")
            end
            posClose = posClose + posOpen
            
            if posOpen+1 == posClose then
              parameters [#parameters+1] = ""
            else
              parameters [#parameters+1] = part:sub (posOpen+1, posClose-1)
            end
            posB = posClose + 1
          end
          
          trailingPart = part:sub (posB)
        end

        if #(command.contents) == 0 then
          newLines [#newLines+1] = leadingPart..trailingPart
          
          return true, true, newLines
          
        else

          if command.paramCount == 0 then
            for cntIdx = 1, #(command.contents) do
              newLines [#newLines+1] = command.contents [cntIdx]
            end
          else
            for cntIdx = 1, #(command.contents) do
              newLines [#newLines+1] = replaceParametersInCommandContent (command.contents [cntIdx], command.paramCount, parameters)
            end
          end
          newLines [1]         = leadingPart..newLines [1]
          newLines [#newLines] = newLines [#newLines]..trailingPart
        
          local finalLines = {}
          
          local bFoundCommand = true
          
          while bFoundCommand do
            finalLines    = {}
            bFoundCommand = false
            
            for newIdx = 1, #newLines do
              local ok, foundCommand, newNewLines = getLinesWithExpandedCommands (lineNumber, fileName, newLines [newIdx])
              
              if not ok then
                return false, false, {}
              end
              
              if foundCommand then
                bFoundCommand = true
              end
              
              for newNewIdx = 1, #newNewLines do
                finalLines [#finalLines+1] = newNewLines [newNewIdx]
              end
            end
            
            if bFoundCommand then
              newLines = finalLines
            end
          end
          
          return true, true, finalLines
          
        end  
      end
    end

    newLines [#newLines+1] = line
    return true, false, newLines
  end
  
  
  function addLineWithExpandedCommandsToResultingLines (lineNumber, fileName, line)
    
    local ok, bDummy, newLines = getLinesWithExpandedCommands (lineNumber, fileName, line)

    if ok then
      for expandedIdx = 1, #newLines do
        resultingLinesOfPreRead [#resultingLinesOfPreRead + 1] = newLines [expandedIdx]
      end
    end
    return ok
  end
  
  
  function lines_from_SPBuchsatz_file (file)
    --print("Einlesen von Datei "..file)
    if not file_exists (file) then 
      print("\n\nFehler: Datei "..file.." nicht gefunden.\n\n")
      return {} 
    end

    resultingLinesOfPreRead = {}
    
    local inAnyTable       = false
    local rawTableLines    = {}
    local idxLineInRawFile = 0

    for lineRaw in io.lines (file) do 
        
      idxLineInRawFile = idxLineInRawFile + 1

      local line = lineRaw
      if line:byte(1) == 0xEF and
         line:byte(2) == 0xBB and 
         line:byte(3) == 0xBF
      then
        -- BOM found, we ignore it
        line = line:sub (4)
      end
                                                             
      local lineStripped = line:gsub(" ", "")

      local bOK, bUserCommandDefinitionFound = parseLineAndStoreUserCommandDefinitionIfAny (idxLineInRawFile, file, line, lineStripped)

      if not bOK then
        print("\n\nSyntax-Fehler in Datei "..file..".\n\n")
        return {}
      end

      if not bUserCommandDefinitionFound then

        if startsWithAnyBeginTableCommand (lineStripped) then
          inAnyTable = true
        end

        if inAnyTable then
          -- Note: rawTableLines will contain all of the table lines
          --       including \\Beginn… and \\End…
          rawTableLines [#rawTableLines + 1] = line
        else
          
          if not addLineWithExpandedCommandsToResultingLines (idxLineInRawFile, file, line) then
            print("\n\nFehler beim Einlesen von Datei "..file..".\n\n")
            return {}
          end
        end

        if startsWithAnyEndTableCommand (lineStripped) then

          local tableLines = preprocessTableLines (rawTableLines)  -- resulting #tableLines will be  >=  #rawTableLines

          for tableIdx = 1, #tableLines do 
            if not (tableLines [tableIdx] == "") then
              addLineWithExpandedCommandsToResultingLines (idxLineInRawFile, file, tableLines [tableIdx])
            end
          end

          rawTableLines = {}
          inAnyTable    = false
        end
      end
    end

    return resultingLinesOfPreRead
  end
  


  function createMainDocument (addExtraLF, einzelneKapitelnummer, einzelneBuchteilnummer, globalNoIndent, skipTrailingPages)
    local weAreInTabular = false
    local mainDocument = io.open("mainDocument.tex","w")
    mainDocument:write("%\n%    This file was automatically created by SPBuchsatz.\n%\n%    Do NOT edit this file:\n%\n%    Your changes would be overwritten when SPBuchsatz is started again.\n%\n%\n%\n%\n%    Diese Datei wurde automatisch von SPBuchsatz erzeugt.\n%\n%    Editiere diese Datei nicht:\n%\n%    Deine AEnderungen wuerden ueberschrieben, wenn SPBuchsatz das naechste Mal gestartet wird.\n%\n\n\n\n\n")

    mainDocument:write("\\evtlSetzeFigurenUndOderOrtslisteZuBeginn\n\n")
    mainDocument:write("% Teile den Skripten mit, dass die Titelei (und evtl. Figuren-/Ortsliste) fertig gesetzt sind.\n")
    mainDocument:write("\\global\\def\\spInternWirSindImHaupttext {x}\n")
    mainDocument:write("\\global\\def\\spInternWirSindImHaupttextOderAufDenSchlussseiten {x}\n")
    mainDocument:write("\\begin{luacode*}\n")
    mainDocument:write("  spInternPagesOfMainPart = true\n")
    mainDocument:write("\\end{luacode*}\n\n")

    local specialFileForRTF = io.open("auxDocumentForRTF.tex","w")
    specialFileForRTF:write("%\n%    This file was automatically created by SPBuchsatz.\n%\n%    Do NOT edit this file:\n%\n%    Your changes would be overwritten when SPBuchsatz is started again.\n%\n%\n%\n%\n%    Diese Datei wurde automatisch von SPBuchsatz erzeugt.\n%\n%    Editiere diese Datei nicht:\n%\n%    Deine AEnderungen wuerden ueberschrieben, wenn SPBuchsatz das naechste Mal gestartet wird.\n%\n%\n%\n%\n%\n")
    
    local wantedChapterIdx = nil
    local wantedChapterIdxLast = nil
    
    
    local einzelneKapitelnummerPrepared = ""

    local einzelneKapitelnummerTrimmed = einzelneKapitelnummer:gsub (" ", "")
    local einzelneKapitelnummerPrepared = einzelneKapitelnummerTrimmed:gsub ("%-%-", "_")
    einzelneKapitelnummerPrepared = einzelneKapitelnummerPrepared:gsub ("_", "-_")
    if startsWith (einzelneKapitelnummerPrepared, "-")
    then
      einzelneKapitelnummerPrepared = "_"..einzelneKapitelnummerPrepared:sub (2)
    end
    
    local gewollteKapitelnummern = split (einzelneKapitelnummerPrepared)
    if (#gewollteKapitelnummern > 0) then
      local ersterToken = gewollteKapitelnummern [1]:gsub ("_", "-")
      wantedChapterIdx = tonumber(ersterToken)
      if #gewollteKapitelnummern > 1 then
        local zweiterToken = gewollteKapitelnummern [2]:gsub ("_", "-")
        wantedChapterIdxLast = tonumber(zweiterToken)
      end
    end
    
    local allChapters = (wantedChapterIdx == nil) or (wantedChapterIdx == -99)
    if allChapters then
      wantedChapterIdx = -99
    else
      if wantedChapterIdxLast == nil then
        print ("\n\n\nVorverarbeitung von Haupttext.tex, aber nur ein Kapitel: "..wantedChapterIdx.."\n")
      else
        print ("\n\n\nVorverarbeitung von Haupttext.tex, aber nur die Kapitel "..wantedChapterIdx.." bis "..wantedChapterIdxLast.."\n")
      end
      currentChapter = 0
    end
    
    local wantedPartIdx = tonumber(einzelneBuchteilnummer)
    local allParts = (wantedPartIdx == nil) or (wantedPartIdx == -99)
    if allParts then
      wantedPartIdx = -99
    else
      print ("\n\n\nVorverarbeitung von Haupttext.tex, aber NUR "..wantedPartIdx..". Buchteil\n")
      currentPart = 0
    end
    
    if allChapters and allParts then
      print ("\n\nVorverarbeitung von Haupttext.tex")
    end

    local fileName = "../tex/Haupttext.tex"

    local rawFileLines = lines_from_SPBuchsatz_file (fileName)
                                                             
                                                             
                                                             
    --print ("Zwischencode-Größe: "..tostring (#rawFileLines).." Zeilen")
                                                             
    for idx = 1, #rawFileLines do
      processPrologChapterStoryCommands (rawFileLines [idx])
    end

    
    local iLine = 0

    if totalNumberOfPrologsNormalChaptersInlineChaptersStories == 0 then
      iLine = iLine+1
      mainDocument:write("\\ErsterAbsatzNachKapitelOderSzeneBeginntrue\\spInternerHookFuerErstzeilenAutomatik\n")
      flagWeAreAtStartOfMaintextAndWeHaveNoChaptersNorStories = true
    end
    
    waitingForClosingOfCommandLevel = 0
    
    io.write ("[")
    for idx = 1, #rawFileLines do
      
      local line, lineStripped, isComment = preprocessLine (rawFileLines [idx])

      if line == "\\DateiHierBeenden" then
        break
      end

      if not isComment then
        iLine = iLine+1
        if (iLine < 10) or (iLine % 10 == 0) then
          io.write (".")
        end
        weAreInTabular = appendFixedLineToMainDocument (line, lineStripped, mainDocument, weAreInTabular, addExtraLF, wantedChapterIdx, wantedChapterIdxLast, wantedPartIdx, globalNoIndent, specialFileForRTF)
      end
    end
    print ("]")

    
    mainDocument:write("\\evtlSetzeFigurenUndOderOrtslisteAmEnde\n\n")
    
    mainDocument:write("\n\n%%%%%%%  Schlussseiten  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n\n")
    mainDocument:write("\\addtocontents{toc}{\\protect\\vspace{0.75\\baselineskip}}\n\n")
    mainDocument:write("\\neueSeite\n\n")
    mainDocument:write("\n\n\\global\\def\\spInternWirSindImHaupttext {}\n\n")
    mainDocument:write("\\KeineSeitenzahl")
    mainDocument:write("\\KeineRBilder")
    mainDocument:write("\\BinImBuchteilfalse")

    mainDocument:write("\\unindentChaptersForPFDTOC\n\n")

    
    if skipTrailingPages then
      print ("\n  *** ACHTUNG: Schlussseiten werden NICHT gesetzt  ***")
    else
      print ("\nVorverarbeitung von Schlussseiten.tex")
      
      local fileName = "../tex/Schlussseiten.tex"

      local rawFileLines = lines_from_SPBuchsatz_file (fileName)
      
      waitingForClosingOfCommandLevel = 0
      
      iLine = 0
      io.write ("[")
      for idx = 1, #rawFileLines do
        
        local line, lineStripped, isComment = preprocessLine (rawFileLines [idx])

        if line == "\\DateiHierBeenden" then
          break
        end
        
        if not isComment then
          iLine = iLine+1
          if (iLine < 10) or (iLine % 10 == 0) then
            io.write (".")
          end
          weAreInTabular = appendFixedLineToMainDocument (line, lineStripped, mainDocument, weAreInTabular, addExtraLF, -99, -99, -99, globalNoIndent, specialFileForRTF)
        end
      end
      print ("]")
    end
    
    io.close (mainDocument)
    io.close (specialFileForRTF)
    print ("\nVorverarbeitung beendet.\n")
  end
\end{luacode*}

                                                             

% - Internal code for creating built-in lists of characters and locations
% - and for preventing hypnenation of character/location names - - - - - - 
                                                             

% Erstellen einer Liste, falls Daten gefunden
% 
% Hauptfunktion:
%     function luaCreateList (keyOfItem, title, listStyle, filename, leadingCommands)
%                                          
% Parameter:
%     #1  Begriff-Schluesselwort (z.B. "Ort:" oder "Figur:")
%     #2  Titel
%     #3  Begrifflisten-Stil ("Klassik", "Eleganz", "Einzug", "Einfach", "Technik")
%     #4  Dateiname
%     #5  Optionfeld, ob VOR der Liste ein \neueSeite auszugeben ist
%
\begin{luacode*}
    function trim(s)
      return s:match'^%s*(.*%S)' or ''
    end
    
    function processLineToAvoidHypenations (keyOfItem, line)
        local posKey = findFirst (line, keyOfItem)
        if posKey > 0 then
            line = trim (line:sub(posKey+keyOfItem:len ()))
            if line:len () == 0 then
                return
            end
            print ("     "..line.."[s]")
            tex.print ("\\hyphenation{"..line.."}")
            tex.print ("\\hyphenation{"..line.."s}")
            tex.print ("")
        end
    end
    
    function luaAvoidHypenations (keyOfItem, filename)
        local listFile = io.open (filename,"r")
        if listFile ~= nil then
            justStartingAvoidHypenations = true
            print ("\n\n *** Silbentrennung wird verhindert für Namen in "..filename.."\n")

            local foundSecondPart = false
            for line in listFile:lines () do
                local lineStripped = line:gsub (" ", "")
                if lineStripped:len () > 0 then
                    if lineStripped == "SPBuchsatzMarkerEndeDerListe" then
                        foundSecondPart = true
                    end
                    if foundSecondPart then
                        local lineNew = ""
                        local posComment = findFirstComment (lineStripped)
                        if posComment == 1 then 
                            goto endOfLoop
                        end
                        posComment = findFirstComment (line)
                        if posComment > 1 then  -- Note: There might be a space between \ and % so we need to have this extra if
                            -- take part before Percent and remove leading/trailing spaces
                            lineNew = line:sub (1, posComment-1):gsub ('^%s*(.-)%s*$', '%1')
                        else
                            lineNew = line:gsub ('^%s*(.-)%s*$', '%1')
                        end
                        processLineToAvoidHypenations (keyOfItem, lineNew)
                    end
                end
                ::endOfLoop::
            end
            print ("")
            io.close (listFile)
        end
    end
    
    justStartingCreateList = true
    vorigerBegriffTitel = ""
    vorigeBegriffZeilen = {}
    vorigeBegriffZeilenZahl = 0
    
    function setzeVorigenBegriff (listStyle, EndReached)
        if vorigerBegriffTitel:len () > 0 then
            local formatierterBegriffTitel = vorigerBegriffTitel
            
            if  not ((listStyle == "Klassik") or (listStyle == "Einfach") or (listStyle == "Einzug")) then
              formatierterBegriffTitel = makeFineFAEs (formatierterBegriffTitel)
            end
            
            if (vorigeBegriffZeilenZahl == 0) or (vorigeBegriffZeilen[1]:len () == 0) then
                tex.print ("\\Begriff{"..formatierterBegriffTitel.."}{}")
                tex.print ("")
            else
              local optionalParameter    = ""
              local formatierteZeileEins = ""
              if startsWith (vorigeBegriffZeilen[1], "|") then
                optionalParameter = "[v]"
                if vorigeBegriffZeilen[1]:len () > 1 then
                  formatierteZeileEins = makeFineFAEs (vorigeBegriffZeilen [1]:sub (2))  -- skip first char
                end
              else
                formatierteZeileEins = makeFineFAEs (vorigeBegriffZeilen [1])
              end
--print("|"..formatierteZeileEins.."|")
              if vorigeBegriffZeilenZahl == 1 then
                tex.print ("\\Begriff"..optionalParameter.."{"..formatierterBegriffTitel.."}{"..formatierteZeileEins.."}")
                tex.print ("")
              else
                tex.print ("\\Begriff"..optionalParameter.."{"..formatierterBegriffTitel.."}{"..formatierteZeileEins)
                tex.print ("")
                if vorigeBegriffZeilenZahl > 2 then
                  for i=2, vorigeBegriffZeilenZahl-1 do                
                    local fineLine = makeFineFAEs (vorigeBegriffZeilen [i])
                    tex.print (fineLine)
                    tex.print ("")
                  end
                end
                tex.print (makeFineFAEs (vorigeBegriffZeilen [vorigeBegriffZeilenZahl]).."}")
                tex.print ("")
              end
            end
        end
        vorigerBegriffTitel = ""
        vorigeBegriffZeilen = {}
        vorigeBegriffZeilenZahl = 0
        if EndReached then
            tex.print ("\\neueSeite")
            tex.print ()
            tex.print ("\\MitSeitenzahl")
            tex.print ()
        end
    end
    
    function processLineToCreateList (keyOfItem, title, listStyle, line, leadingCommands)
        local lineStripped = line:gsub(" ", "")
        local posComment = findFirstComment (lineStripped)
        if posComment > 0 then
            if posComment == 1 then 
                return
            end
            posComment = findFirstComment (line)
            if posComment > 0 then  -- Note: There might be a space between \ and % so we need to have this extra if
              line = line:sub (1, posComment-1)
            end
        end
        if lineStripped:len () == 0 then 
            return
        end
        
        
        local foundPageBreak = startsWith (lineStripped, "\\neueSeite")

        local foundEndeBegriff = startsWith(lineStripped,"\\EndeBegriff")

        local foundHSpace    = startsWith (lineStripped, "\\BegrifflisteAbstandOben")
                            or startsWith (lineStripped, "\\Linie")
                            or startsWith (lineStripped, "\\ViertelLeerzeileInListe")
                            or startsWith (lineStripped, "\\HalbeLeerzeileInListe")
                            or startsWith (lineStripped, "\\DreiViertelLeerzeileInListe")
                            or startsWith (lineStripped, "\\LeerzeileInListe")
                            or startsWith (lineStripped, "\\FaktorLeerzeileInListe")

        local foundImage     = startsWith (lineStripped, "\\BildBreite")
                            or startsWith (lineStripped, "\\BildHoehe")
                            or startsWith (lineStripped, "\\BildAbfallendOben")
                         
        local foundKey = startsWith (lineStripped, keyOfItem)
        
        if foundKey or foundEndeBegriff or foundPageBreak or foundHSpace or foundImage then
            if justStartingCreateList then
                justStartingCreateList = false
                if leadingCommands == "x" then
                    tex.print ("\\neueSeite")
                    tex.print("\\global\\def\\spInternWirSindImHaupttext {}")
                end
                tex.print ("\\KeineSeitenzahl")
                tex.print ()
                print ("\n\n *** Setzen der Liste: "..title.."\n")
                
                local formatierterTitel = title
                if not ((listStyle == "Klassik") or (listStyle == "Tabelle")) then
                  formatierterTitel = makeFineFAEs (formatierterTitel)
                end
                
                tex.print ("\\Begriffliste{"..listStyle.."}{"..formatierterTitel.."}")
                tex.print ("")
            end
            
            setzeVorigenBegriff (listStyle, false)
            
            if foundPageBreak then
                tex.print ("\\clearpage")
            else 
                if foundHSpace or foundImage then
                    tex.print ("\\LeerzeileOben")
                    tex.print ("\\GeheZeileNachOben")
                    tex.print (lineStripped)
                else
                    if foundKey then
                        local posKey = findFirst (line, keyOfItem)
                        line = trim (line:sub(posKey+keyOfItem:len ()))
                        if line:len () == 0 then
                            return
                        end
                        print ("     "..line)
                        vorigerBegriffTitel = line
                    end
                end
            end
        else
            if vorigerBegriffTitel == "" then
                tex.print (line)
                tex.print ("")
            else
                if not justStartingCreateList then
                    vorigeBegriffZeilenZahl = vorigeBegriffZeilenZahl + 1
                    vorigeBegriffZeilen [vorigeBegriffZeilenZahl] = trim (line)
                end
            end
        end
    end
    
    function luaCreateList (keyOfItem, title, listStyle, filename, leadingCommands)
        local listFile = io.open (filename,"r")
        if listFile ~= nil then
            justStartingCreateList = true
            for line in listFile:lines () do
                local lineStripped = line:gsub(" ", "")
                if lineStripped:len () > 0 then
                    if lineStripped == "SPBuchsatzMarkerEndeDerListe" then
                    
                        setzeVorigenBegriff (listStyle, true)
                    
                        io.close (listFile)
                        return
                    end
                    processLineToCreateList (keyOfItem, title, listStyle, line, leadingCommands)
                end
            end
            setzeVorigenBegriff (listStyle, true)
            io.close (listFile)
        end
    end
\end{luacode*}



% - Internal code for auto-loading and registering fonts - - - - - - - - -


% Gegebenfalls im lokalen Verzeichis ../meine_Fonts/ liegende Schriften
% laden und Befehle für sie erzeugen:
\newcommand{\registriereSchriftDateien}[1]{%
\luaexec{
function FolderExists(entry)
  local attr=lfs.attributes (entry)
  if attr and (attr.mode == "directory") then
    return true
  else
    return false
  end
end
function firstFileInFolder(fileName)
         print(fileName)
  if fileName:len () == 0 then
    return false;
  end
  if not FolderExists (fileName) then
    return ""
  end 
  for entry in lfs.dir (fileName) do
    if entry ~= "." and entry ~= ".." and entry ~= ".DS_Store" and entry ~= "info.txt" then
      local fullEntry=fileName.."/"..entry
      local attr=lfs.attributes (fullEntry)
      if attr.mode == "file" then
        return fullEntry
      end
    end
  end
  return ""
end
function ersteDatei(firstFilename, otherFilename)
  local retValue = firstFileInFolder (firstFilename)
  if retValue:len() == 0 then
    retValue = firstFileInFolder (otherFilename)
  end
  return retValue
end
function warningFontIgnored(message,fontname)
  tex.print("\\PackageWarning{SPBuchsatz}"..
            "{Schrift-Ordner '"..fontname.."' wird ignoriert:"..
            "\\MessageBreak "..message.."}")
end
function ladeSchrift(dir2,fontname)
  print("")
  print("Ordner gefunden: "..fontname)
  if fontname=="NeueSchrift" then
    print(" Dies ist der bei Bedarf von dir zu kopierende Beispiel-Ordner.")
    print(" Schrift-Ordner ignoriert:", fontname)
    print("")
    return
  end
  if fontname:match("\%A") then
    warningFontIgnored("Umlaute oder Ziffern sind verboten, erlaubt sind nur große oder kleine Buchstaben von A-Z.",fontname)
    print("")
    return
  end
  local c1=string.sub(fontname, 1, 1)
  if not (c1:match("\%u")) then
    warningFontIgnored("Name beginnt nicht mit einem Großbuchstaben.",fontname)
    print("")
    return
  end
  local fileNormal =      ersteDatei(dir2.."/Normal",        dir2.."/Regular")
  local fileKursiv =      ersteDatei(dir2.."/Kursiv",        dir2.."/Italic")
  local fileFett =        ersteDatei(dir2.."/Fett",          dir2.."/Bold")
  local fileFettUndKursiv=ersteDatei(dir2.."/FettUndKursiv", dir2.."/BoldItalic")
  local DateiExt=string.sub(fileNormal, -3)
  print("  Normal:        "..fileNormal)
  print("  Kursiv:        "..fileKursiv)
  print("  Fett:          "..fileFett)
  print("  FettUndKursiv: "..fileFettUndKursiv)
  print("  Typ:           "..DateiExt)
  if DateiExt:len () == 3 then
    tex.print ("\\typeout{Laden von \\protect\\Schrift"..fontname.."\\space und Erstellen der Befehle:}")
    tex.print ("\\typeout{\\space\\space\\space\\space\\protect\\BeginnSchrift"..fontname.."}")
    tex.print ("\\typeout{\\space\\space\\space\\space\\protect\\EndeSchrift"..fontname.."}")
    tex.print ("\\typeout{\\space\\space\\space\\space\\protect\\schrift"..fontname.."{...}}")
    local strNewFontFamilyCommand = "\\newfontfamily\\Schrift"..fontname.."[Extension=."..DateiExt
    if (fileNormal and fileNormal:len () > 4)
    then
      fileNormal        = string.sub(fileNormal,        1,fileNormal:len ()-4)
      strNewFontFamilyCommand = strNewFontFamilyCommand..",UprightFont="..fileNormal
    end
    if (fileKursiv and fileKursiv:len () > 4)
    then
      fileKursiv        = string.sub(fileKursiv,        1,fileKursiv:len ()-4)
      strNewFontFamilyCommand = strNewFontFamilyCommand..",ItalicFont="..fileKursiv
    end
    if (fileFett and fileFett:len () > 4) 
    then
      fileFett          = string.sub(fileFett,          1,fileFett:len ()-4)
      strNewFontFamilyCommand = strNewFontFamilyCommand..",BoldFont="..fileFett
    end
    if (fileFettUndKursiv and fileFettUndKursiv:len () > 4) 
    then
      fileFettUndKursiv = string.sub(fileFettUndKursiv, 1,fileFettUndKursiv:len ()-4)
      strNewFontFamilyCommand = strNewFontFamilyCommand..",BoldItalicFont="..fileFettUndKursiv
    end
    strNewFontFamilyCommand = strNewFontFamilyCommand..",Ligatures=TeX]{"..fontname.."}"
    tex.print (strNewFontFamilyCommand)
    tex.print ("\\newcommand{\\schrift"..fontname.."}[1]{\\begin{Schrift"..fontname.."}"..string.char(35).."1\\end{Schrift"..fontname.."}}")
    tex.print ("\\def\\BeginnSchrift"..fontname.."{\\begin{Schrift"..fontname.."}}")
    tex.print ("\\def\\EndeSchrift"..fontname.."{\\end{Schrift"..fontname.."}}")
    tex.print ("\\typeout{}")
  else
    warningFontIgnored("Datei(en) nicht gefunden.",fontname)
    print("")
  end
end
function sucheSchriften(dir)
  assert(dir and dir ~= "", "directory parameter is missing or empty")
  if string.sub(dir, -1) == "/" then
    dir=string.sub(dir, 1, -2)
  end
  print("")
  print("Suchen nach weiteren Schriften in "..dir)
  for entry in lfs.dir(dir) do
    if (entry ~= ".") and (entry ~= "..") then
      local pfad=dir.."/"..entry
      local attr=lfs.attributes(pfad)
      if attr.mode == "directory" then
        ladeSchrift(pfad,entry)
      end
    end
  end
end
sucheSchriften ("#1")
}}

                                                             
                                                             
% - Internal code to pre-process text fields of the central document - - -

\begin{luacode*}
  function insertParAfterBlockStartEndCommands (line)
    local pos = findFirst (line, "\\Beginn")
    if pos == 0 then
      pos = findFirst (line, "\\Ende")
    end
    
    if pos == 0 then
      return line
    end
    
    pos = findAfter (line, " ", pos)
    
    if pos == 0 then
      return line:sub (1, pos+1) .. insertParAfterBlockStartEndCommands (line:sub (8))
    end
    
    local partA = line:sub (1, pos-1)
    local partB = line:sub (pos)
    local partBStripped = partB:gsub(" ", "")
    
    if partBStripped:len () == 0 then
      return line
    end
    
    if not (partA [pos-1] == "*") then
      if startsWith (partBStripped, "*") then
        local posStar = findFirst (partB, "*")
        partA = partA .. partB:sub (1, posStar)
        partB = partB:sub (posStar + 1)
      end
    end

    partBStripped = partB:gsub(" ", "")
    
    if (not startsWith (partBStripped, "\\par ")) and
       (not startsWith (partBStripped, "[")) and
       (not startsWith (partBStripped, "{"))
    then
      partA = partA .. " \\par "
    end
    
    return partA .. insertParAfterBlockStartEndCommands (partB)
  end
  
  
  function getLinesFromVorverarbeitungTmp ()
    
    local fineLines = {}
    
    local rawFile = io.open ("Vorverarbeitung.tmp","r")
    if rawFile ~= nil then
      local iLine = 0
      for line in rawFile:lines () do
        local lineStripped = line:gsub(" ", "")
        if iLine == 0 then
          iLine = iLine + 1
          print ("  \\"..lineStripped)
          fineLines [iLine] = "\\def\\" .. lineStripped .. "{"
        else
          if lineStripped:len () > 0 then 

            local fixedLine = insertParAfterBlockStartEndCommands (line)
            
            iLine = iLine + 1
            fineLines [iLine] = makeFineFAEs (fixedLine)
          end
        end
      end
      io.close (rawFile)

      if iLine > 1 then
        fineLines [1] = fineLines [1] .. "%"
      end
      fineLines [iLine] = fineLines [iLine] .. "}"
    end
    
    return fineLines
  end
  
  
  function redefineFieldFromVorverarbeitungTmp ()
    local lines = getLinesFromVorverarbeitungTmp ()
    if (#lines > 0) then
      tex.print (lines)
    end
  end
\end{luacode*}
                                                             
% Hilfsbefehl zum Schreiben eines Feldes in eine temp. Datei und
% dann Aufruf der Vorverarbeitung per Lua:
\newcommand{\spInternVorverarbeitungFuerFeld}[2]{%
\newwrite\tmpFile
\immediate\openout\tmpFile Vorverarbeitung.tmp
\immediate\write\tmpFile{#1}
\immediate\write\tmpFile{\expandonce#2}
\immediate\closeout\tmpFile
\luaexec{redefineFieldFromVorverarbeitungTmp ()}
}

\newcommand{\spSetzeHilfsdateiDerFeldVorverarbeitungZurueck}{%
\newwrite\tmpFile
\immediate\openout\tmpFile Vorverarbeitung.tmp
\immediate\closeout\tmpFile
}


\begin{luacode*}
  function printGridLines (lineHeight, totalHeight)
    local lineCount = math.floor (totalHeight / lineHeight)
    local y = -0.125
    for i=0, lineCount+1 do
      tex.print ("\\put(-1,"..tostring (y).."){\\color{\\spColorOfGridLines}\\line(1,0){\\LenToUnit{\\dimexpr \\layerwidth+2pt}}}")
      y = y + lineHeight
    end
    y = y - lineHeight
    tex.print ("\\put(-0.85,0){\\color{\\spColorOfGridLines}\\line(0,1){\\LenToUnit{\\dimexpr "..tostring (y).."pt}}}")
    tex.print ("\\put(\\LenToUnit{\\dimexpr \\layerwidth+0.85pt},0){\\color{\\spColorOfGridLines}\\line(0,1){\\LenToUnit{\\dimexpr "..tostring (y).."pt}}}")
  end
\end{luacode*}



% - end of file - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
